#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import sys


def xy2rp(x, y):
    """Convert (x,y) to (R,phi) in polar coordinate

    Args:
        x (float): x coordinate
        y (float): y coordinate

    Returns:
        R (float): radius
        phi (float): angle in rad
    """
    R = np.sqrt(x ** 2 + y ** 2)
    if x > 0.0 and y >= 0.0:  # [0,pi/2)
        phi = np.arcsin(y / R)
    elif x <= 0.0 and y > 0.0:  # [pi/2, pi)
        phi = np.arccos(x / R)
    elif x < 0.0 and y <= 0.0:  # [pi, 3/2 pi)
        phi = np.arccos(-x / R) + np.pi
    elif x >= 0.0 and y < 0.0:  # [3/2 pi, 2pi)
        phi = np.arcsin(y / R) + 2 * np.pi
    else:
        raise ValueError("Something wrong with your inputs ({:f}, {:f}).".format(x, y))
    return R, phi


def map_matrix(xx, first=True, second=True):
    """Map matrix to be complete (closed)
    Arguments:
      xx -- 2D numpy array
      first -- boolean, default: True, if increase the first dimension
      second -- boolean, default: True, if increase the second dimension

    Returns:
      new -- the new matrix with dimension increased
    """
    a, b = np.shape(xx)
    # only first
    if first and not second:
        new = np.zeros((a + 1, b))
        new[0:a, 0:b] = xx[0:a, 0:b]
        new[a, 0:b] = xx[0, 0:b]
    # only second
    elif not first and second:
        new = np.zeros((a, b + 1))
        new[0:a, 0:b] = xx[0:a, 0:b]
        new[0:a, b] = xx[0:a, 0]
    # both direction
    elif first and second:
        new = np.zeros((a + 1, b + 1))
        new[0:a, 0:b] = xx[0:a, 0:b]
        new[a, 0:b] = xx[0, 0:b]
        new[0:a, b] = xx[0:a, 0]
        new[a, b] = xx[0, 0]
    # otherwise return the original matrix
    else:
        return xx
    return new


def toroidal_period(vec, nfp=1):
    """
    vec: [x,y,z] data
    Nfp: =1, toroidal number of periodicity
    """
    phi = 2 * np.pi / nfp
    vec = np.atleast_2d(vec)
    new_vec = vec.copy()
    for ifp in range(nfp):
        if ifp == 0:
            continue
        rotate = np.array(
            [
                [np.cos(ifp * phi), np.sin(ifp * phi), 0],
                [-np.sin(ifp * phi), np.cos(ifp * phi), 0],
                [0, 0, 1],
            ]
        )
        new_vec = np.concatenate((new_vec, np.matmul(vec, rotate)))
    return new_vec


def print_progress(
    iteration, total, prefix="Progress", suffix="Complete", decimals=1, bar_length=60
):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        bar_length  - Optional  : character length of bar (Int)
    """
    str_format = "{0:." + str(decimals) + "f}"
    percents = str_format.format(100 * (iteration / float(total)))
    filled_length = int(round(bar_length * iteration / float(total)))
    bar = "█" * filled_length + "-" * (bar_length - filled_length)

    sys.stdout.write("\r%s |%s| %s%s %s" % (prefix, bar, percents, "%", suffix)),

    if iteration == total:
        sys.stdout.write("\n")
    sys.stdout.flush()
    return


# Smart way to check where to plot
def get_figure(axes=None, **kwargs):
    """
    Check where to plot
    Parameters:
        axes: matplotlib.pyplot axis, axis to plot (default: None)
        kwargs: keyword arguments
    Return:
        f, ax
        f  : matplotlib.pyplot figure
        ax : matplotlib.pyplot axis
    """
    import matplotlib.pyplot as plt

    if axes is None:
        # No axes provided
        f, axes = plt.subplots(**kwargs)
        """
        f = plt.gcf()
        if len(f.axes):
            # normal situation in which existing figures should be respected and left alone
            f, axes = plt.subplots(**kwargs)
        else:
            #  made a empty figure for using
            axes = f.add_subplot(**kwargs)
         """
    else:
        # axes = np.atleast_1d(axes)
        f = axes.get_figure()
    return f, axes


def colorbar(mappable, **kwargs):
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    import matplotlib.pyplot as plt

    last_axes = plt.gca()
    ax = mappable.axes
    fig = ax.figure
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar = fig.colorbar(mappable, cax=cax, **kwargs)
    plt.sca(last_axes)
    return cbar


def kwargs2dict(**kwargs):
    return kwargs


def vmecMN(mpol, ntor):
    # manipulate VMEC index
    mn = (2 * ntor + 1) * mpol - ntor  # total number of Fourier harmonics
    xm = np.zeros((mn,), dtype=int)
    xn = np.zeros((mn,), dtype=int)
    imn = 0
    for ii in range(mpol):
        for jj in range(-ntor, ntor + 1):
            if ii == 0 and jj < 0:
                continue
            xm[imn] = ii
            xn[imn] = jj
            imn += 1
    return xm, xn


def trigfft(y, tr=-1):
    """calculate trigonometric coefficients using FFT
    Assuming the periodicity is 2*pi
    params:
        y -- 1D array for Fourier transformation
        tr -- Truncation number (default: -1)
    return:
        a dict containing
        'n' -- index
        'rcos' -- cos coefficients of the real part
        'rsin' -- sin coefficients of the real part
        'icos' -- cos coefficients of the imag part
        'isin' -- sin coefficients of the imag part
    """
    from scipy.fftpack import fft

    N = len(y)
    if N % 2 == 0:  # even
        half = N // 2 - 1
        end = half + 2
    else:
        half = (N - 1) // 2
        end = half + 1
    assert tr <= end, "Truncation number should be smaller than dimension!"
    comp = fft(y) / N
    a_k = np.zeros(end, dtype=np.complex)
    b_k = np.zeros(end, dtype=np.complex)
    a_k[0] = comp[0]
    for n in range(1, half + 1):
        a_k[n] = comp[n] + comp[N - n]
        b_k[n] = (comp[n] - comp[N - n]) * 1j
    if N % 2 == 0:  # even number
        a_k[end - 1] = comp[N // 2]
    index = np.arange(end)

    return {
        "n": index[:tr],
        "rcos": np.real(a_k[:tr]),
        "rsin": np.real(b_k[:tr]),
        "icos": np.imag(a_k[:tr]),
        "isin": np.imag(b_k[:tr]),
    }


def trigfft2(y):
    """calculate trigonometric coefficients using FFT
    Assuming the periodicity is 2*pi
    params:
        y -- 2D array for Fourier transformation
    return:
        a dict containing
        'n' -- 1D array, n index
        'm' -- 1D array, m index
        'rcos' -- 2D array, cos coefficients of the real part
        'rsin' -- 2D array, sin coefficients of the real part
        'icos' -- 2D array, cos coefficients of the imag part
        'isin' -- 2D array, sin coefficients of the imag part
    """
    from scipy.fftpack import fft2, fftshift

    M, N = y.shape
    mn = M * N
    comp = fft2(y) / mn
    if M % 2 == 0:  # even
        half = M // 2 - 1
        end = half + 2
    else:
        half = (M - 1) // 2
        end = half + 1
    if N % 2 == 0:  # even
        mid0 = N // 2
        start = 1
        nmin = -N // 2
        nmax = N // 2 - 1
    else:
        mid0 = (N - 1) // 2
        start = 0
        nmin = -(N - 1) // 2
        nmax = (N - 1) // 2
    a_k = np.zeros((end, N), dtype=np.complex)
    b_k = np.zeros((end, N), dtype=np.complex)
    # find mapping
    a_k[0, 0] = comp[0, 0]
    for n in range(1, N):
        a_k[0, n] = comp[0, n] + comp[0, N - n]
        b_k[0, n] = (comp[0, n] - comp[0, N - n]) * 1j
    for m in range(1, (M + 1) // 2):
        a_k[m, 0] = comp[m, 0] + comp[M - m, 0]
        b_k[m, 0] = (comp[m, 0] - comp[M - m, 0]) * 1j
        for n in range(1, N):
            a_k[m, n] = comp[m, n] + comp[M - m, N - n]
            b_k[m, n] = (comp[m, n] - comp[M - m, N - n]) * 1j
    if M % 2 == 0 and N % 2 == 0:  # even
        a_k[end - 1, N // 2] = comp[M // 2, N // 2]
    a_k = fftshift(a_k, axes=1)
    b_k = fftshift(b_k, axes=1)
    a_k[0, start:mid0] = 0 + 1j * 0
    b_k[0, start:mid0] = 0 + 1j * 0
    mm = np.arange(end)
    nn = np.arange(nmin, nmax + 1)
    return {
        "n": nn,
        "m": mm,
        "rcos": np.real(a_k),
        "rsin": np.real(b_k),
        "icos": np.imag(a_k),
        "isin": np.imag(b_k),
    }


def fft_deriv(y):
    from scipy.fftpack import fft, ifft

    N = len(y)
    comp = fft(y)
    if N % 2 == 0:
        dt = (
            np.arange(N)
            - np.concatenate((np.zeros(N // 2), [N // 2], N * np.ones(N // 2 - 1)))
        ) * 1j
    else:
        dt = (
            np.arange(N) - np.concatenate((np.zeros(N // 2), N * np.ones(N // 2 + 1)))
        ) * 1j
    return ifft(comp * dt)


def trig2real(theta, zeta=None, xm=[], xn=[], fmnc=None, fmns=None):
    """Trigonometric coefficients to real space points

    Args:
        theta (numpy.ndarray): Theta values to be evaluated.
        zeta (numpy.ndarray, optional): Zeta values to be evaluated if discretizing in 2D. Defaults to None.
        xm (list, optional): Poloidal Fourier modes. Defaults to [].
        xn (list, optional): Toroidal Fourier modes. Defaults to [].
        fmnc ([type], optional): Cosine Fourier coefficients Defaults to None.
        fmns ([type], optional): Sin Fourier coefficients. Defaults to None.

    Returns:
        numpy.ndarray: The discretized values in real space.
    """
    if zeta is None:
        return _trig2real_1d(theta, xm, fmnc, fmns)
    else:
        return _trig2real_2d(theta, zeta, xm, xn, fmnc, fmns)


def _trig2real_1d(theta, xm, fmnc=None, fmns=None):
    _mt = np.reshape(xm, (-1, 1)) * theta
    _cos = np.cos(_mt)
    _sin = np.sin(_mt)
    f = np.zeros((1, len(theta)))
    if fmnc is not None:
        f += np.matmul(np.reshape(fmnc, (1, -1)), _cos)
    if fmns is not None:
        f += np.matmul(np.reshape(fmns, (1, -1)), _sin)
    return f.ravel()


def _trig2real_2d(theta, zeta, xm, xn, fmnc=None, fmns=None):
    npol, ntor = len(theta), len(zeta)
    _tv, _zv = np.meshgrid(theta, zeta, indexing="ij")
    # mt - nz (in matrix)
    _mtnz = np.matmul(np.reshape(xm, (-1, 1)), np.reshape(_tv, (1, -1))) - np.matmul(
        np.reshape(xn, (-1, 1)), np.reshape(_zv, (1, -1))
    )
    _cos = np.cos(_mtnz)
    _sin = np.sin(_mtnz)

    f = np.zeros((1, npol * ntor))
    if fmnc is not None:
        f += np.matmul(np.reshape(fmnc, (1, -1)), _cos)
    if fmns is not None:
        f += np.matmul(np.reshape(fmns, (1, -1)), _sin)
    return f.reshape(npol, ntor)


def real2trig_2d(f, xm, xn, theta, zeta):
    """Fourier decomposition in 2D

    Args:
        f (numpy.ndarray): The 2D function to be decomposed. Size: [npol, ntor].
        xm (numpy.ndarray): Poloildal mode number. Size: [mn,]
        xn (numpy.ndarray): Toroildal mode number. Size: [mn,]
        theta (numpy.ndarray): Poloidal angles. Size:[npol,].
        zeta (numpy.ndarray): Toroidal angles. Size:[ntor,]

    Returns:
        numpy.ndarray, numpy.ndarray: Cos harmonics, sin harmonics. Size: [mn,]
    """
    npol, ntor = len(theta), len(zeta)
    assert (npol, ntor) == np.shape(
        f
    ), "F function dimension should be consistent with theta, zeta."
    _tv, _zv = np.meshgrid(theta, zeta, indexing="ij")
    # mt - nz (in matrix)
    _mtnz = np.matmul(np.reshape(xm, (-1, 1)), np.reshape(_tv, (1, -1))) - np.matmul(
        np.reshape(xn, (-1, 1)), np.reshape(_zv, (1, -1))
    )
    _cos = np.cos(_mtnz)
    _sin = np.sin(_mtnz)

    fmnc = np.ravel(np.matmul(_cos, f.reshape(-1, 1)))
    fmns = np.ravel(np.matmul(_sin, f.reshape(-1, 1)))
    fac = 2.0 / (npol * ntor)
    # m=0, n=0 term or m=0 terms?
    ind = np.logical_and(xm == 0, xn == 0)
    fmnc[ind] *= 0.5
    fmns[ind] *= 0.5
    return fmnc * fac, fmns * fac


def vmec2focus(
    vmec_file,
    focus_file="plasma.boundary",
    bnorm_file="",
    ns=-1,
    curpol=1.0,
    flipsign=False,
):
    """Prepare FOCUS input boundary

    Args:
        vmec_file (str): VMEC input or output filename.
        focus_file (str, optional): FOCUS boundary filename to be written. Defaults to 'plasma.boundary'.
        bnorm_file (str, optional): BNORM output filename. Defaults to ''.
        ns (int, optional): VMEC surface index. Defaults to -1.
        curpol (float, optional): Normalization factor related to poloidal current. Defaults to 1.0.
        flipsign (bool, optional): Bool value to flip the sign of Bn coefficients. Defaults to False.
    """
    # check VMEC format
    if "wout_" in vmec_file:
        import xarray

        wout = xarray.open_dataset(vmec_file)
        rmnc = wout["rmnc"].values
        zmns = wout["zmns"].values
        rbc = rmnc[ns, :]
        zbs = zmns[ns, :]
        # non-stellarator-symmetric terms
        if int(wout["lasym__logical__"].values):
            rmns = wout["rmns"].values
            zmnc = wout["zmnc"].values
            rbs = rmns[ns, :]
            zbc = zmnc[ns, :]
        else:
            rbs = np.zeros_like(rbc)
            zbc = np.zeros_like(zbs)
        nfp = int(wout["nfp"].values)
        xm = np.array(wout["xm"], dtype=int)
        xn = np.array(wout["xn"], dtype=int) // nfp
        curpol = 2.0 * np.pi / nfp * wout["rbtor"].values
    elif "input." in vmec_file:
        import f90nml

        nml = f90nml.read(vmec_file)
        indata = nml["indata"]
        mpol = indata["mpol"]
        ntor = indata["ntor"]
        nfp = indata["nfp"]
        arr_rbc = np.array(indata["rbc"])
        arr_zbs = np.array(indata["zbs"])
        arr_rbc[arr_rbc == None] = 0
        arr_zbs[arr_zbs == None] = 0
        try:
            arr_rbs = np.array(indata["rbs"])
            arr_zbc = np.array(indata["zbc"])
            arr_rbs[arr_rbs == None] = 0
            arr_zbc[arr_zbc == None] = 0
        except KeyError:
            arr_rbs = np.zeros_like(arr_rbc)
            arr_zbc = np.zeros_like(arr_rbc)
        nmin, mmin = indata.start_index["rbc"]
        mlen, nlen = np.shape(indata["rbc"])
        xm = []
        xn = []
        rbc = []
        zbs = []
        rbs = []
        zbc = []
        for i in range(mlen):
            m = i + mmin
            if m > mpol:
                continue
            for j in range(nlen):
                n = j + nmin
                if n > ntor:
                    continue
                xm.append(m)
                xn.append(n)
                rbc.append(arr_rbc[i, j])
                zbs.append(arr_zbs[i, j])
                rbs.append(arr_rbs[i, j])
                zbc.append(arr_zbc[i, j])
    else:
        raise FileExistsError(
            "Please check your argument. Should be VMEC input or output!"
        )
    # parse BNORM output if necessary
    if bnorm_file == "":
        Nbnf = 0
    else:
        bm = []
        bn = []
        bns = []
        # bnc = []
        with open(bnorm_file, "r") as bfile:
            for line in bfile:
                tmp = line.split()  # BNORM format: m n Bn_sin
                bm.append(int(tmp[0]))
                bn.append(int(tmp[1]))
                bns.append(float(tmp[2]))
        Nbnf = len(bm)
        bnc = np.zeros(Nbnf)
        bns = np.array(bns) * curpol
        if flipsign:
            bns *= -1
    # write FOCUS input
    mn = len(xm)
    with open(focus_file, "w") as fofile:
        fofile.write("# bmn   bNfp   nbf " + "\n")
        fofile.write("{:d} \t {:d} \t {:d} \n".format(mn, nfp, Nbnf))
        fofile.write("#plasma boundary" + "\n")
        fofile.write("# n m Rbc Rbs Zbc Zbs" + "\n")
        for i in range(mn):
            fofile.write(
                "{:4d}  {:4d} \t {:15.7E}  {:15.7E}  {:15.7E}  {:15.7E} \n".format(
                    xn[i], xm[i], rbc[i], rbs[i], zbc[i], zbs[i]
                )
            )
        fofile.write(
            "#Bn harmonics curpol= {:15.7E} ; I_p={:15.7E} A. \n".format(
                curpol, curpol * nfp / (2 * np.pi) * 5e6
            )
        )
        fofile.write("# n m bnc bns \n")
        for i in range(Nbnf):
            fofile.write(
                "{:d} \t {:d} \t {:15.7E} \t {:15.7E} \n".format(
                    -bn[i], bm[i], bnc[i], bns[i]
                )
            )
            # FOCUS uses mu - nv
    return


def booz2focus(booz_file, ns=-1, focus_file="plasma.boundary", tol=1e-6, Nfp=1):
    """convert BOOZ_XFORM output into FOCUS format plasma surface (in Boozer coordinates)

    Args:
        booz_file (str): Netcdf file of BOOZ_XFORM output
        ns (int, optional): The specific flux surface you want to convert. Defaults to -1.
        focus_file (str, optional): FOCUS plasma boundary filename. Defaults to 'plasma.boundary'.
        tol ([type], optional): Tolerance to truncate. Defaults to 1E-6.
        Nfp (int, optional): [description]. Defaults to 1.
    """
    import xarray

    booz = xarray.open_dataset(booz_file)
    mn = int(booz["mnboz_b"].values)
    xm = np.array(booz["ixm_b"])
    xn = np.array(booz["ixn_b"]) / Nfp
    rbc = np.array(booz["rmnc_b"][ns, :])
    # rbs = np.zeros(mn)
    zbs = np.array(booz["zmns_b"][ns, :])
    # zbc = np.zeros(mn)
    pmns = np.array(booz["pmns_b"][ns, :])
    # pmnc = np.zeros(mn)

    # Nfp = 1
    Nbnf = 0

    amn = 0
    for imn in range(mn):
        if (abs(rbc[imn]) + abs(zbs[imn] + abs(pmns[imn]))) > tol:
            amn += 1  # number of nonzero coef.
    with open(focus_file, "w") as fofile:
        fofile.write("# bmn   bNfp   nbf " + "\n")
        fofile.write("{:d} \t {:d} \t {:d} \n".format(amn, Nfp, Nbnf))
        fofile.write("#plasma boundary" + "\n")
        fofile.write("# n m Rbc Rbs Zbc Zbs Pmnc Pmns" + "\n")
        for imn in range(mn):
            if (abs(rbc[imn]) + abs(zbs[imn] + abs(pmns[imn]))) > tol:
                fofile.write(
                    "{:4d}  {:4d} \t {:23.15E}  {:12.5E}  {:12.5E}  {:23.15E}  {:12.5E}  {:23.15E} \n".format(
                        int(xn[imn]),
                        xm[imn],
                        rbc[imn],
                        0.0,
                        0.0,
                        zbs[imn],
                        0.0,
                        pmns[imn],
                    )
                )
        fofile.write("#Bn harmonics \n")
        fofile.write("# n m bnc bns" + "\n")
    print("Finished write FOCUS input file at ", focus_file)
    return


def read_focus_boundary(filename):
    """Read FOCUS/FAMUS plasma boundary file

    Args:
        filename (str): File name and path.

    Returns:
        boundary (dict): Dict contains the parsed data.
            nfp : number of toroidal periods
            nfou : number of Fourier harmonics for describing the boundary
            nbn : number of Fourier harmonics for Bn
            surface : Toroidal surface dict, containing 'xm', 'xn', 'rbc', 'rbs', 'zbc', 'zbs'
            bnormal : Input Bn dict, containing 'xm', 'xn', 'bnc', 'bns'
    """
    boundary = {}
    surf = {}
    bn = {}
    with open(filename, "r") as f:
        line = f.readline()  # skip one line
        line = f.readline()
        num = int(line.split()[0])  # harmonics number
        nfp = int(line.split()[1])  # number of field periodicity
        nbn = int(line.split()[2])  # number of Bn harmonics
        boundary["nfp"] = nfp
        boundary["nfou"] = num
        boundary["nbn"] = nbn
        # read boundary harmonics
        xm = []
        xn = []
        rbc = []
        rbs = []
        zbc = []
        zbs = []
        line = f.readline()  # skip one line
        line = f.readline()  # skip one line
        for i in range(num):
            line = f.readline()
            line_list = line.split()
            n = int(float(line_list[0]))
            m = int(float(line_list[1]))
            xm.append(m)
            xn.append(n)
            rbc.append(float(line_list[2]))
            rbs.append(float(line_list[3]))
            zbc.append(float(line_list[4]))
            zbs.append(float(line_list[5]))
        surf["xm"] = np.array(xm)
        surf["xn"] = np.array(xn)
        surf["rbc"] = np.array(rbc)
        surf["rbs"] = np.array(rbs)
        surf["zbc"] = np.array(zbc)
        surf["zbs"] = np.array(zbs)
        boundary["surface"] = surf
        # read Bn fourier harmonics
        xm = []
        xn = []
        bnc = []
        bns = []
        if nbn > 0:
            line = f.readline()  # skip one line
            line = f.readline()  # skip one line
        for i in range(nbn):
            line = f.readline()
            line_list = line.split()
            n = int(float(line_list[0]))
            m = int(float(line_list[1]))
            xm.append(m)
            xn.append(n)
            bnc.append(float(line_list[2]))
            bns.append(float(line_list[3]))
        bn["xm"] = np.array(xm)
        bn["xn"] = np.array(xn)
        bn["bnc"] = np.array(bnc)
        bn["bns"] = np.array(bns)
        boundary["bnormal"] = bn
    return boundary


def write_focus_boundary(filename, surf, nfp=1, bn=None, **kwargs):
    """Write the Fourier harmonics down in FOCUS format

    Args:
        filename (str): File name to be saved.
        surf (dict): Plasma surface information, containing 'xm', 'xn', 'rbc', 'rbs', 'zbc', zbs'.
        nfp (int, optional): Number of field periodicity. Defaults to 1.
        bn (dict, optional): Nonzero Bn information, containing 'xm', 'xn', 'bnc', 'bns'. Defaults to None.
    """
    # write Fourier coefficients
    mn = len(surf["xn"])
    try:
        nbn = len(bn["xn"])
    except TypeError:
        nbn = 0
    # write data
    with open(filename, "w") as fofile:
        fofile.write("# bmn   bNfp   nbf " + "\n")
        fofile.write("{:d} \t {:d} \t {:d} \n".format(mn, nfp, nbn))
        fofile.write("#plasma boundary" + "\n")
        fofile.write("# n m Rbc Rbs Zbc Zbs" + "\n")
        for imn in range(mn):
            fofile.write(
                "{:4d}  {:4d} \t {:23.15E}  {:23.15E}  {:23.15E}  {:23.15E} \n".format(
                    surf["xn"][imn],
                    surf["xm"][imn],
                    surf["rbc"][imn],
                    surf["rbs"][imn],
                    surf["zbc"][imn],
                    surf["zbs"][imn],
                )
            )
        fofile.write("#Bn harmonics \n")
        fofile.write("# n m bnc bns \n")
        if nbn > 0:
            for imn in range(nbn):
                fofile.write(
                    "{:4d}  {:4d} \t {:23.15E}  {:23.15E}  \n".format(
                        bn["xn"][imn],
                        bn["xm"][imn],
                        bn["bnc"][imn],
                        bn["bns"][imn],
                    )
                )
    return


def div0(a, b):
    return np.divide(a, b, out=np.zeros_like(a), where=b != 0)


def biot_savart(pos, xyz, current, dxyz=None):
    from coilpy_fortran import hanson_hirshman, biot_savart

    if dxyz is None:
        # no tangent provided
        return hanson_hirshman(pos, xyz, current)
    else:
        # tangent provided
        return biot_savart(pos, xyz, current, dxyz)


def rotation_matrix(alpha=0.0, beta=0.0, gamma=0.0):
    """A genera rotation matrix using yaw, pitch, and roll angles

    Args:
        alpha (float, optional): The yaw angle (rotating around the z-axis). Defaults to 0.0.
        beta (float, optional): The pitch angle (rotating around the y-axis). Defaults to 0.0.
        gamma (float, optional): The roll angle (rotating around the x-axis). Defaults to 0.0.

    Returns:
        [type]: [description]
    """
    ca = np.cos(alpha)
    sa = np.sin(alpha)
    cb = np.cos(beta)
    sb = np.sin(beta)
    cc = np.cos(gamma)
    sc = np.sin(gamma)
    return np.array(
        [
            [ca * cb, ca * sb * sc - sa * cc, ca * sb * cc + sa * sc],
            [sa * cb, sa * sb * sc + ca * cc, sa * sb * cc - ca * sc],
            [-sb, cb * sc, cb * cc],
        ]
    )



import h5py
import os  # for path.abspath
import keyword  # for getting python keywords

# Modified from J. Schilling (jonathan.schilling@ipp.mpg.de)'s script for SPEC


class HDF5:
    """Create a python object for a HDF5 file.

    Returns:
        HDF5 class: python HDF5 object.

    Use as s = HDF5(filename), e.g. s=HDF5("ext.h5")
    This class can be iterated or entered.
    To check all the items, you can use `self.inventory()`.
    """

    def __init__(self, *args, **kwargs):
        """Constructor

        Args:
            arg[0] (str): The name of a file or an item inside the root object.
                          If args[0] is not a filename, kwargs['content'] should be the
                          content to be added as self.`args[0]`.
        """

        _content = None
        if kwargs.get("content") is None:
            # assume arg[0] is a filename
            _content = h5py.File(args[0], "r")

            # keep track of which file this object corresponds to
            self.filename = os.path.abspath(args[0])
        elif isinstance(kwargs["content"], h5py.Group):
            _content = kwargs["content"]

        if _content is not None:
            for key in _content:
                if isinstance(_content[key], h5py.Group):
                    # recurse into group
                    setattr(self, key, HDF5(content=_content[key]))
                elif isinstance(_content[key], h5py.Dataset):  # read dataset
                    if (
                        key in keyword.kwlist
                    ):  # add underscore avoiding assigning python keywords
                        setattr(self, key + "_", _content[key][()])
                    else:
                        try:  # this should be simplified when FOCUS writes the correct format
                            if len(_content[key][()]) == 1:
                                # if just one element, use the value directly
                                setattr(self, key, _content[key][0])
                            else:  # arrays
                                setattr(self, key, _content[key][()])
                        except TypeError:  # scalar
                            setattr(self, key, _content[key][()])

        if isinstance(_content, h5py.File):
            _content.close()

    # needed for iterating over the contents of the file
    def __iter__(self):
        return iter(self.__dict__)

    def __next__(self):
        return next(self.__dict__)

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        return

    def inventory(self, prefix=""):
        """Print a list of items contained in this object

        Args:
            prefix (str, optional): Header to be printed. Defaults to "".
        """
        _prefix = ""
        if prefix != "":
            _prefix = prefix + "/"

        for a in self:
            try:
                # recurse into member
                getattr(self, a).inventory(prefix=_prefix + a)
            except AttributeError:
                # print item name
                print(_prefix + a)



from .hdf5 import HDF5
from .misc import get_figure, map_matrix
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import FuncFormatter


class FOCUSHDF5(HDF5):
    """FOCUS output hdf5 file"""

    # initialization, test = FOCUSHDF5('focus_test.h5')
    def __init__(self, filename, periodic=False, **kwargs):
        """Initialization

        Keyword arguments:
            filenmae -- string, path and name to FOCUS output hdf5 file,
                        usually in the format of 'focus_*.h5'
            periodic -- logical, map all 2D surface data automatically,
                        (default: True)
        """
        # read data
        super().__init__(filename)
        # print version
        try:
            abc = ""
            for i in self.version:
                abc += i.decode("utf-8")
            self.version = abc
            print("FOCUS version: " + self.version)
        except AttributeError:
            print(filename + " is not a valid FOCUS output. Please check.")
            raise
        # add additional colume and row for plotting
        if periodic:
            self.xsurf = map_matrix(self.xsurf)
            self.ysurf = map_matrix(self.ysurf)
            self.zsurf = map_matrix(self.zsurf)
            self.nx = map_matrix(self.nx)
            self.ny = map_matrix(self.ny)
            self.nz = map_matrix(self.nz)
            self.nn = map_matrix(self.nn)
            self.Bx = map_matrix(self.Bx)
            self.By = map_matrix(self.By)
            self.Bz = map_matrix(self.Bz)
            self.Bn = map_matrix(self.Bn)
            self.plas_Bn = map_matrix(self.plas_Bn)
        return

    # convergence plot
    def convergence(self, term="bnorm", iteration=True, axes=None, **kwargs):
        # get figure
        fig, axes = get_figure(axes)
        # set default plotting parameters
        kwargs["linewidth"] = kwargs.get("linewidth", 2.5)  # line width
        # kwargs['marker'] = kwargs.get('marker', 'o') # extent
        # get iteration data
        if iteration:
            abscissa = 1 + np.arange(self.iout)
            _xlabel = "iteration"
        else:
            abscissa = self.evolution[0, :]  # be careful; DF is not saving wall-time
            _xlabel = "wall time [Second]"
        # plot data
        if term.lower() == "chi":
            data = self.evolution[1, :]
            kwargs["label"] = kwargs.get("label", r"$\chi^2$")  # line label
        elif term.lower() == "gradient":
            data = self.evolution[2, :]
            kwargs["label"] = kwargs.get(
                "label", r"$|d \chi^2 / d {\bf X}|$"
            )  # line label
        elif term.lower() == "bnorm":
            data = self.evolution[3, :]
            kwargs["label"] = kwargs.get("label", r"$f_{B_n}$")  # line label
        elif term.lower() == "bharm":
            data = self.evolution[4, :]
            kwargs["label"] = kwargs.get("label", r"$f_{B_{mn}}$")  # line label
        elif term.lower() == "tflux":
            data = self.evolution[5, :]
            kwargs["label"] = kwargs.get("label", r"$f_{\Psi}$")  # line label
        elif term.lower() == "ttlen":
            data = self.evolution[6, :]
            kwargs["label"] = kwargs.get("label", r"$f_L$")  # line label
        elif term.lower() == "cssep":
            data = self.evolution[7, :]
            kwargs["label"] = kwargs.get("label", r"$f_{CS}$")  # line label
        elif term.lower() == "curv":
            data = self.evolution[8, :]
            kwargs["label"] = kwargs.get("label", r"$f_{curv}$")  # line label
        elif term.lower() == "all":
            lines = []
            lines.append(
                self.convergence(
                    term="chi",
                    iteration=iteration,
                    axes=axes,
                    linestyle="-",
                    color="k",
                    **kwargs
                )
            )
            lines.append(
                self.convergence(
                    term="gradient",
                    iteration=iteration,
                    axes=axes,
                    linestyle="--",
                    color="k",
                    **kwargs
                )
            )
            lines.append(
                self.convergence(
                    term="bnorm",
                    iteration=iteration,
                    axes=axes,
                    linestyle="-.",
                    color="r",
                    **kwargs
                )
            )
            lines.append(
                self.convergence(
                    term="tflux",
                    iteration=iteration,
                    axes=axes,
                    linestyle="-.",
                    color="g",
                    **kwargs
                )
            )
            lines.append(
                self.convergence(
                    term="bharm",
                    iteration=iteration,
                    axes=axes,
                    linestyle="-.",
                    color="b",
                    **kwargs
                )
            )
            lines.append(
                self.convergence(
                    term="ttlen",
                    iteration=iteration,
                    axes=axes,
                    linestyle="-.",
                    color="c",
                    **kwargs
                )
            )
            lines.append(
                self.convergence(
                    term="cssep",
                    iteration=iteration,
                    axes=axes,
                    linestyle="-.",
                    color="m",
                    **kwargs
                )
            )
            lines.append(
                self.convergence(
                    term="curv",
                    iteration=iteration,
                    axes=axes,
                    linestyle="-.",
                    color="y",
                    **kwargs
                )
            )
            # fig.legend(loc='upper right', frameon=False, ncol=2, prop={'size':16})
            plt.legend()
            return lines
        else:
            raise ValueError("unsupported option for term")
        line = axes.semilogy(abscissa, data, **kwargs)
        axes.tick_params(axis="both", which="major", labelsize=15)
        axes.set_xlabel(_xlabel, fontsize=15)
        axes.set_ylabel("cost functions", fontsize=15)
        # fig.legend(loc='upper right', frameon=False, prop={'size':24, 'weight':'bold'})
        plt.legend()
        return line

    # poincare plot
    def poincare_plot(self, color=None, prange="full", **kwargs):
        """Poincare plot from FOCUS output.
        Args:
             color (matplotlib color, or None): dot colors. Defaults to None (rainbow).
             prange (str, optional): Plot range, one of ["upper", "lower", "all"]. Defaults to "full".
             kwargs : matplotlib scatter keyword arguments

        Returns:
             None
        """
        import matplotlib.pyplot as plt
        from matplotlib import cm

        # get figure and ax data
        if plt.get_fignums():
            fig = plt.gcf()
            ax = plt.gca()
        else:
            fig, ax = plt.subplots()

        # get colors
        if color is None:
            colors = cm.rainbow(np.linspace(1, 0, self.pp_ns))
        else:
            colors = [color] * self.pp_ns
        kwargs["s"] = kwargs.get("s", 0.1)  # dotsize
        # scatter plot
        for i in range(self.pp_ns):
            # determine whether plot upper or lower
            if prange == "upper":
                cond = self.ppz[:, i] > 0
            elif prange == "lower":
                cond = self.ppz[:, i] < 0
            else:
                cond = np.ones_like(self.ppz[:, i], dtype=bool)
            ax.scatter(
                self.ppr[:, i][cond], self.ppz[:, i][cond], color=colors[i], **kwargs
            )
        plt.axis("equal")
        plt.xlabel("R [m]", fontsize=20)
        plt.ylabel("Z [m]", fontsize=20)
        plt.xticks(fontsize=16)
        plt.yticks(fontsize=16)
        return

    # Bnorm plot
    def Bnorm(self, plottype="2D", source="all", axes=None, flip=False, **kwargs):
        """Plot Bn distribution.

        Keyword arguments:
            plottype -- string, '2D' (default) or '3D', determine the plottype
            source -- string, 'coil', 'plasma', 'sum' or 'all' (default), data source
            axes  -- matplotlib.pyplot or mayavi.mlab axis, axis to be
                     plotted on  (default None)
            flip -- logical, determine how to calculate Bn from coil,
                    True: coil_Bn = Bn - plas_Bn; False: coil_Bn = Bn + plas_Bn

        Returns:
           obj -- matplotlib.pyplot or mayavi.mlab plotting object
        """
        obj = []
        # check coil_Bn
        if flip:
            coil_Bn = self.Bn - self.plas_Bn
        else:
            coil_Bn = self.Bn + self.plas_Bn
        # 2D plots
        if plottype.lower() == "2d":
            # prepare coordinates
            ntheta = self.Nteta
            nzeta = self.Nzeta
            nt, nz = self.Bn.shape
            if self.IsSymmetric == 0:
                zeta_end = 2 * np.pi
            elif self.IsSymmetric == 1:
                zeta_end = 2 * np.pi / self.Nfp
            elif self.IsSymmetric == 2:
                zeta_end = np.pi / self.Nfp
            else:
                raise ValueError("Something wrong with IsSymmetric.")

            def theta(x, pos):
                return "{:3.2f}".format(np.pi / ntheta + x * (2 * np.pi) / ntheta)

            def zeta(x, pos):
                return "{:3.2f}".format(0.5 * zeta_end / nzeta + x * zeta_end / nzeta)

            theta_format = FuncFormatter(theta)
            zeta_format = FuncFormatter(zeta)
            # planar plotting
            if source.lower() == "all":
                fig, axes = get_figure(axes, ncols=3, sharex=True)
                axes = np.atleast_1d(axes)
                # kwargs['aspect'] = kwargs.get('aspect', nt/(2.0*nz*self.Nfp))
                kwargs["aspect"] = kwargs.get("aspect", "auto")
            else:
                fig, axes = get_figure(axes)
                axes = np.atleast_1d(axes)
                kwargs["aspect"] = kwargs.get("aspect", float(nt / nz))
            # prepare axes
            for ax in axes:
                ax.xaxis.set_major_formatter(zeta_format)
                ax.yaxis.set_major_formatter(theta_format)
            # set default plotting parameters
            kwargs["cmap"] = kwargs.get("cmap", "RdBu_r")  # colormap
            kwargs["origin"] = kwargs.get("origin", "lower")  # number of contours
            kwargs["extent"] = kwargs.get("extent", [0, nt, 0, nz])  # extent
            plt.subplots_adjust(hspace=0.05)
            # imshow
            if source.lower() == "coil":
                obj.append(axes[0].imshow(np.transpose(coil_Bn), **kwargs))
                axes[0].set_title("Bn from coils", fontsize=15)
                axes[0].set_ylabel(r"$\theta$", fontsize=14)
                axes[0].set_xlabel(r"$\phi$", fontsize=14)
            elif source.lower() == "plasma":
                obj.append(axes[0].imshow(np.transpose(self.plas_Bn), **kwargs))
                axes[0].set_title("Bn from plasma", fontsize=15)
                axes[0].set_ylabel(r"$\theta$", fontsize=14)
                axes[0].set_xlabel(r"$\phi$", fontsize=14)
            elif source.lower() == "sum":
                obj.append(axes[0].imshow(np.transpose(self.Bn), **kwargs))
                axes[0].set_title("Residual Bn", fontsize=15)
                axes[0].set_ylabel(r"$\theta$", fontsize=14)
                axes[0].set_xlabel(r"$\phi$", fontsize=14)
            elif source.lower() == "all":
                vmin = np.min([self.plas_Bn, self.Bn, coil_Bn])
                vmax = np.max([self.plas_Bn, self.Bn, coil_Bn])
                obj.append(
                    axes[0].imshow(
                        np.transpose(self.plas_Bn), vmin=vmin, vmax=vmax, **kwargs
                    )
                )
                # plt.setp(axes[0].get_xticklabels(), visible=False)
                axes[1].set_title(
                    "Bn from plasma (left), coil (mid), overall(right).", fontsize=14
                )
                obj.append(
                    axes[1].imshow(
                        np.transpose(coil_Bn), vmin=vmin, vmax=vmax, **kwargs
                    )
                )
                # axes[1].set_title('Bn from coils', fontsize=15)
                plt.setp(axes[1].get_yticklabels(), visible=False)
                obj.append(
                    axes[2].imshow(
                        np.transpose(self.Bn), vmin=vmin, vmax=vmax, **kwargs
                    )
                )
                plt.setp(axes[2].get_yticklabels(), visible=False)
                # axes[2].set_title('Residual Bn', fontsize=15)
                axes[0].set_ylabel(r"$\theta$", fontsize=14)
                # axes[1].set_ylabel(r'$\theta$', fontsize=14)
                # axes[2].set_ylabel(r'$\theta$', fontsize=14)
                axes[0].set_xlabel(r"$\phi$", fontsize=14)
                axes[1].set_xlabel(r"$\phi$", fontsize=14)
                axes[2].set_xlabel(r"$\phi$", fontsize=14)
            else:
                raise ValueError("unsupported option for source")
            fig.subplots_adjust(right=0.85)
            cbar_ax = fig.add_axes([0.86, 0.10, 0.04, 0.8])
            fig.colorbar(obj[0], cax=cbar_ax)
        return obj

    # Bmod plot
    def Bmod(self):
        return

    def plot(
        self, engine="plotly", scalars="bn", fig=None, ax=None, show=True, **kwargs
    ):
        """Plot 3D field on the plasma from FOCUS ouput

        Args:
            engine (str, optional): 3D plot engine, one of ["mayavi", "pyplot", "plotly"]. Defaults to "plotly".
            scalars (str, optional): Scalar function on the surface, one of ["Bn", "B", "plas_Bn"] or 2D arrays. Defaults to "bn".
            fig (, optional): Figure to be plotted on. Defaults to None.
            ax (, optional): Axis to be plotted on. Defaults to None.
            show (bool, optional): If show the plotly figure immediately. Defaults to True.
            kwargs: optional keyword arguments for plotting.
        Raises:
            ValueError: Plot engine should be one value of ["mayavi", "pyplot", "plotly"].
        """
        full = True if self.IsSymmetric == 0 else False
        xsurf = map_matrix(self.xsurf, first=full)
        ysurf = map_matrix(self.ysurf, first=full)
        zsurf = map_matrix(self.zsurf, first=full)
        label = "plasma"
        if scalars.lower() == "bn":
            scalars = self.Bn
            label = "Bn"
        elif scalars.lower() == "plas_bn":
            scalars = self.plas_Bn
            label = "Bn_plasma"
        elif scalars.lower() == "b":
            scalars = np.sqrt(self.Bx ** 2 + self.By ** 2 + self.Bz ** 2)
            label = "|B|"
        scalars = map_matrix(scalars, first=full)
        if engine == "pyplot":
            import matplotlib.pyplot as plt

            # plot in matplotlib.pyplot
            if ax is None or ax.name != "3d":
                fig = plt.figure()
                ax = fig.add_subplot(111, projection="3d")
            kwargs.setdefault("cmap", "coolwarm")
            ax.plot_surface(xsurf, ysurf, zsurf, c=scalars, **kwargs)
            ax.set_title(label)
            fig.colorbar()
        elif engine == "mayavi":
            # plot 3D surface in mayavi.mlab
            from mayavi import mlab  # to overrid plt.mlab

            kwargs.setdefault("colormap", "coolwarm")
            mlab.mesh(xsurf, ysurf, zsurf, scalars=scalars, **kwargs)
            mlab.colorbar(title=label)
        elif engine == "plotly":
            import plotly.graph_objects as go

            kwargs.setdefault("colorbar", go.surface.ColorBar(title=label))
            if fig is None:
                fig = go.Figure()
            fig.add_trace(
                go.Surface(x=xsurf, y=ysurf, z=zsurf, surfacecolor=scalars, **kwargs)
            )
            fig.update_layout(scene_aspectmode="data")
            if show:
                fig.show()
        else:
            raise ValueError("Invalid engine option {pyplot, mayavi, noplot}")

    # write vtk
    def toVTK(self, name=None, full=False, **kwargs):
        """Save surface and magnetic field data into VTK file
        Arguments:
          name -- string, VTK file name. default: None, if None, using self.filename
          full -- boolean, if save the entire torus
          **kwargs -- external data will be saved in VTK file

        Return:
          VTK file name
        """
        from pyevtk.hl import gridToVTK

        # automatically get a file name, focus_*.h5 -> vtk_*
        if name is None:
            name = self.filename[:-3].replace("focus_", "vtk_")
        xx = self.xsurf
        yy = self.ysurf
        zz = self.zsurf
        Bx = self.Bx
        By = self.By
        Bz = self.Bz
        Bn = self.Bn
        plas_Bn = self.plas_Bn
        toroidal = full
        xx = np.atleast_3d(map_matrix(xx, first=toroidal))
        yy = np.atleast_3d(map_matrix(yy, first=toroidal))
        zz = np.atleast_3d(map_matrix(zz, first=toroidal))
        Bn = np.atleast_3d(map_matrix(Bn, first=toroidal))
        plas_Bn = np.atleast_3d(map_matrix(plas_Bn, first=toroidal))
        B = (
            np.atleast_3d(map_matrix(Bx, first=toroidal)),
            np.atleast_3d(map_matrix(By, first=toroidal)),
            np.atleast_3d(map_matrix(Bz, first=toroidal)),
        )
        if full:
            pass
        data = {"Bn": Bn, "plas_Bn": plas_Bn, "B": B}
        data.update(kwargs)
        return gridToVTK(name, xx, yy, zz, pointData=data)

    def curvature(
        self,
        iteration=True,
        vlines=True,
        shift_ind=0,
        icoil=1,
        NS=128,
    ):
        """Curvature ploting for the FOCUS-spline paper [arXiv:2107.02123] by N. Lonigro

        Args:
            iteration (bool, optional): [description]. Defaults to True.
            vlines (bool, optional): [description]. Defaults to True.
            shift_ind (int, optional): [description]. Defaults to 0.
            icoil (int, optional): [description]. Defaults to 1.
            NS (int, optional): [description]. Defaults to 128.

        Returns:
            [type]: [description]
        """
        import math

        # get figure
        max_ind = -1
        min_ind = -1
        # get iteration data
        if iteration:
            abscissa = np.arange(NS)
            _xlabel = "Segment number"
        else:
            abscissa = self.evolution[0, :]  # be careful; DF is not saving wall-time
            _xlabel = "wall time [Second]"
        # get data
        data = object.__getattribute__(self, "curvature_{:}".format(icoil))
        data_s = object.__getattribute__(self, "straight_{:}".format(icoil))

        if (data_s[0] == 0) and (data_s[NS - 1] != 0) and (data_s[k + 1] == 0):
            max_ind = 0
        if (data_s[0] == 0) and (data_s[1] != 0) and (data_s[k - 1] == 0):
            min_ind = 0
        if (data_s[NS - 1] == 0) and (data_s[NS - 2] != 0) and (data_s[0] == 0):
            max_ind = NS - 1
        if (data_s[NS - 1] == 0) and (data_s[0] != 0) and (data_s[NS - 2] == 0):
            min_ind = NS - 1
        for k in range(1, len(data_s) - 2):
            if (data_s[k] == 0) and (data_s[k - 1] != 0) and (data_s[k + 1] == 0):
                max_ind = np.max([max_ind, k])
            if (
                (data_s[k] == 0)
                and (data_s[k + 1] != 0)
                and (data_s[k - 1] == 0)
                and min_ind != -1
            ):
                min_ind = np.min([min_ind, k])
            if (
                (data_s[k] == 0)
                and (data_s[k + 1] != 0)
                and (data_s[k - 1] == 0)
                and min_ind == -1
            ):
                min_ind = k
        if max_ind > min_ind and shift_ind == 0:
            shift_ind = math.floor(0.5 * (NS - max_ind - min_ind))
        if max_ind < min_ind and shift_ind == 0:
            shift_ind = math.floor(0.5 * NS + 0.5 * (NS - max_ind - min_ind))
        return [shift_ind, max_ind - min_ind]


import numpy as np


class Mgrid(object):
    def __init__(self, r, z, phi, Br, Bz, Bphi, nfp=1):
        self.r = r
        self.z = z
        self.phi = phi
        self.Br = Br
        self.Bz = Bz
        self.Bphi = Bphi
        self.nfp = nfp
        self.interpolate()
        return

    @classmethod
    def read_mgrid_bin(cls, filename, extcur=None):

        f = open(filename, "rb")
        # nr, nz, etc.
        (nbytes,) = np.fromfile(f, dtype="i4", count=1)
        assert nbytes == 20, "Integer was not written as 'i4'. Try i{:}".format(
            nbytes // 5
        )
        nr, nz, nphi, nfp, nextcur = np.fromfile(f, dtype="i4", count=5)
        if nextcur < 0:
            style_2000 = True
            nextcur = abs(nextcur)
        else:
            style_2000 = False
        (nbytes,) = np.fromfile(f, dtype="i4", count=1)
        # rmin, rmax, etc.
        (nbytes,) = np.fromfile(f, dtype="i4", count=1)
        assert nbytes == 32, "Real was not written as 'f8'. Try f{:}".format(
            nbytes // 4
        )
        rmin, zmin, rmax, zmax = np.fromfile(f, dtype="f8", count=4)
        (nbytes,) = np.fromfile(f, dtype="i4", count=1)
        # curlabel
        (ncurlabel,) = np.fromfile(f, dtype="i4", count=1) // 30
        assert ncurlabel == nextcur, "ncurlabel != nextcur"
        curlabel = np.fromfile(f, dtype="S30", count=nextcur)
        (nbytes,) = np.fromfile(f, dtype="i4", count=1)
        # Br, Bp, Bz
        nrpz = 3 * nr * nphi * nz
        Br = []
        Bp = []
        Bz = []
        for i in range(nextcur):
            (nbytes,) = np.fromfile(f, dtype="i4", count=1)
            assert nbytes == nrpz * 8, "B was not written as 'f8'. Try f{:}".format(
                nbytes // nrpz
            )
            B = np.fromfile(f, dtype="f8", count=nrpz)
            if style_2000:
                Bvec = B.reshape((nphi, nz, nr, 3)).T
                Br.append(Bvec[0])
                Bp.append(Bvec[1])
                Bz.append(Bvec[2])
            else:
                Bvec = B.reshape((nphi, nz, nr, 3)).T
                Br.append(Bvec[0])
                Bz.append(Bvec[1])
                Bp.append(Bvec[2])
            (nbytes,) = np.fromfile(f, dtype="i4", count=1)
        # mgrid_mode, "S" or "N"
        if style_2000:
            (nbytes,) = np.fromfile(f, dtype="i4", count=1)
            (mgrid_mode,) = np.fromfile(f, dtype="S1", count=1)
            (nbytes,) = np.fromfile(f, dtype="i4", count=1)
            # raw currents
            (nbytes,) = np.fromfile(f, dtype="i4", count=1)
            assert (
                nbytes == 8 * nextcur
            ), "Real was not written as 'f8'. Try f{:}".format(nbytes // nextcur)
            extcur = np.fromfile(f, dtype="f8", count=nextcur)
        else:
            mgrid_mode = "N"
        f.close()
        # read extcur if needed
        if extcur is None:
            if not style_2000:  # FOCUS/FAMUS old format
                extcur = np.ones(nextcur)
        elif isinstance(extcur[0], str):  # read from file
            extcur = np.genfromtxt(extcur)
        # generate coordinates
        rr = np.linspace(rmin, rmax, nr)
        zz = np.linspace(zmin, zmax, nz)
        phi = np.linspace(0, 2 * np.pi / nfp, nphi + 1)
        # sum data
        Br_total = np.zeros((nr, nz, nphi + 1))
        Bz_total = np.zeros((nr, nz, nphi + 1))
        Bp_total = np.zeros((nr, nz, nphi + 1))
        for i in range(nextcur):
            Br_total[:, :, :-1] += Br[i] * extcur[i]
            Bz_total[:, :, :-1] += Bz[i] * extcur[i]
            Bp_total[:, :, :-1] += Bp[i] * extcur[i]
        # add one additional toroidal cross-section
        Br_total[:, :, -1] = Br_total[:, :, 0]
        Bz_total[:, :, -1] = Bz_total[:, :, 0]
        Bp_total[:, :, -1] = Bp_total[:, :, 0]
        return cls(
            r=rr, z=zz, phi=phi, Br=Br_total, Bz=Bz_total, Bphi=Bp_total, nfp=nfp
        )

    def interpolate(self, **kwargs):
        from scipy.interpolate import RegularGridInterpolator

        self.Br_func = RegularGridInterpolator(
            (self.r, self.z, self.phi), self.Br, bounds_error=False, fill_value=0
        )
        self.Bz_func = RegularGridInterpolator(
            (self.r, self.z, self.phi), self.Bz, bounds_error=False, fill_value=0
        )
        self.Bphi_func = RegularGridInterpolator(
            (self.r, self.z, self.phi), self.Bphi, bounds_error=False, fill_value=0
        )
        return

    def bfield(self, rzp):
        rzp[2] = rzp[2] % (2 * np.pi / self.nfp)
        Br = self.Br_func(rzp)
        Bz = self.Bz_func(rzp)
        Bphi = self.Bphi_func(rzp)
        return np.array([Br, Bz, Bphi])

    def tracing(self, r0, z0, phi0=0.0, niter=100, **kwargs):
        from scipy.integrate import solve_ivp
        from .misc import print_progress

        def fieldline(phi, rz):
            rzp = np.array([rz[0], rz[1], phi])
            B = np.ravel(self.bfield(rzp))
            return [B[0] / B[2] * rz[0], B[1] / B[2] * rz[0]]

        dphi = 2 * np.pi / self.nfp
        phi = phi0 + dphi * np.arange(niter)
        nlines = len(r0)
        lines = []
        for i in range(nlines):
            points = [[r0[i], z0[i]]]
            for j in range(niter):
                print_progress(i * niter + j, nlines * niter)
                phi_start = phi[j]
                sol = solve_ivp(
                    fieldline,
                    (phi_start, phi_start + dphi),
                    points[j],
                    method="LSODA",
                    **kwargs
                )
                points.append(sol.y[:, -1])
            lines.append(np.array(points))
        return lines

    def poincare_plot(self, r0, z0, phi0=0.0, niter=100, color=None, **kwargs):
        import matplotlib.pyplot as plt
        from matplotlib import cm

        # get figure and ax data
        if plt.get_fignums():
            fig = plt.gcf()
            ax = plt.gca()
        else:
            fig, ax = plt.subplots()

        data = self.tracing(r0, z0, phi0, niter)
        ns = len(data)
        # get colors
        if color is None:
            colors = cm.rainbow(np.linspace(1, 0, ns))
        else:
            colors = [color] * ns
        kwargs["s"] = kwargs.get("s", 0.1)  # dotsize
        # scatter plot
        for i in range(ns):
            ax.scatter(data[i][:, 0], data[i][:, 1], color=colors[i], **kwargs)
        plt.axis("equal")
        plt.xlabel("R [m]", fontsize=20)
        plt.ylabel("Z [m]", fontsize=20)
        plt.xticks(fontsize=16)
        plt.yticks(fontsize=16)
        return


from coilpy import Coil
import numpy as np

# read
ellipse = Coil.read_makegrid("ellipse.coils")
assert ellipse.num == 16, "Coil number is read incorrectly!"
assert len(ellipse.data[0].x) == 129, "Segment number is read incorrectly!"
assert ellipse.data[15].I == -1e6, "Coil current is read incorrectly!"
assert ellipse.data[10].group == 3, "Coil group is read incorrectly!"

# plot
ellipse.plot(irange=range(0, 16, 4))
ellipse.plot(irange=range(0, 16, 4), enginer="plotly", plot2d=True)

# save as VTK files
# lines
ellipse.toVTK("ellipse")
# finite-build
label = []
for icoil in list(ellipse):
    zsign = icoil.z[:-1] > 0
    label += zsign.astype(int).tolist()
ellipse.toVTK(
    "ellipse.vtk", line=False, width=0.05, height=0.05, cell_data={"z_sign": [label]}
)

# calculate B field
b = np.array([-5.85704462e-04, 2.94453517e-03, -1.63013362e-18])
assert np.allclose(ellipse.data[0].bfield([0, 0, 0]), b)

# misc
ellipse.data[1].interpolate()
ellipse.data[1].magnify(ratio=2.0)

# save
ellipse.save_makegrid("test.coils")



"""
This is a python package for plotting and data processing in stellarator optimization.

It contains several python functions on HDF5, Fourier surfaces, FOCUS, cois, 
magnetic dipoles, STELLOPT, VMEC, BOOZ_XFORM, etc.

The repository is available at https://github.com/zhucaoxiang/CoilPy.

For full documenation, please check https://zhucaoxiang.github.io/CoilPy/api/coilpy.html.
"""
__version__ = "0.3.13"

# local packages
from .misc import colorbar, get_figure, kwargs2dict, map_matrix
from .misc import print_progress, toroidal_period, vmecMN, xy2rp
from .misc import trigfft, fft_deriv, trig2real, vmec2focus
from .misc import real2trig_2d, booz2focus, read_focus_boundary, div0
from .misc import biot_savart, rotation_matrix
from .hdf5 import HDF5
from .surface import FourSurf
from .dipole import Dipole
from .focushdf5 import FOCUSHDF5
from .coils import Coil, SingleCoil
from .stellopt import STELLout
from .vmec import VMECout
from .booz_xform import BOOZ_XFORM
from .mgrid import Mgrid

# from coilpy_fortran import hanson_hirshman, biot_savart



# modified from stellopt.pySTEL

from __future__ import print_function, absolute_import
from builtins import map, filter, range
from .sortedDict import SortedDict
import numpy as np
import matplotlib.pyplot as plt
from .misc import get_figure

__all__ = ["STELLout", "OMFITascii"]


class OMFITascii(object):
    """
    OMFIT class used to interface with ASCII files

    :param filename: filename passed to OMFITobject class

    :param fromString: string that is written to file

    :param \**kw: keyword dictionary passed to OMFITobject class
    """

    def __init__(self, filename, **kw):
        fromString = kw.pop("fromString", None)
        self.filename = filename
        if fromString is not None:
            with open(self.filename, "wb") as f:
                f.write(fromString.encode("utf-8"))

    def read(self):
        """
        Read ASCII file and return content

        :return: string
        """
        return open(self.filename, "r").read()

    def write(self, value):
        """
        Write string value to ASCII file

        :param value: string to be written to file

        :return: string
        """
        open(self.filename, "w").write(value)
        return value

    def append(self, value):
        """
        Append string value to ASCII file

        :param value: string to be written to file

        :return: string
        """
        open(self.filename, "a").write(value)
        return self.read()


class STELLout(SortedDict, OMFITascii):
    """
    OMFITobject used to interface with stellopt.* file in STELLOPT outputs.

    :param filename: filename passed to OMFITascii class

    All additional key word arguments passed to OMFITascii

    """

    def __init__(self, filename, **kwargs):
        SortedDict.__init__(self)
        OMFITascii.__init__(self, filename, **kwargs)
        self.dynaLoad = True

    def load(self):
        """Load the file and parse it into a sorted dictionary"""
        file_handle = open(self.filename, "r")
        niter = 0
        for line in file_handle:
            if "ITER" in line:
                niter = niter + 1
            if "MIN" in line:
                niter = niter - 1
        self["ITER"] = np.ndarray((niter, 1))
        file_handle.seek(0)
        line = file_handle.readline()
        ttype, wh = line.split()
        self[ttype] = float(wh)

        # Enter Loop
        citer = -1
        while True:
            line = file_handle.readline()
            if line == "":
                break
            ttype, hw = line.split(" ", 1)
            if ttype == "ITER":
                if "MIN" in hw:
                    break
                citer = citer + 1
                self[ttype][citer] = int(hw)
                continue
            else:
                h, w = hw.split()
                h = int(h)
                w = int(w)
                line = file_handle.readline()
            if ttype not in self:
                self[ttype] = np.ndarray((niter, h, w))
            for i in range(h):
                line = file_handle.readline()
                val = np.fromstring(line, sep=" ")
                self[ttype][citer, i, :] = val
        file_handle.close()
        for item in list(self):
            # print(item)
            if "VERSION" == item:
                continue
            elif "ITER" == item:
                continue
            elif item in [
                "ASPECT",
                "ASPECT_MAX",
                "BETA",
                "CURTOR",
                "PHIEDGE",
                "VOLUME",
                "WP",
                "RBTOR",
                "R0",
                "Z0",
                "BETATOR",
                "BETAPOL",
            ]:
                self[item + "_target"] = np.array(self[item][:, :, 0])
                self[item + "_sigma"] = np.array(self[item][:, :, 1])
                self[item + "_equil"] = np.array(self[item][:, :, 2])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
            elif item == "BALLOON":
                self[item + "_target"] = np.array(self[item][:, :, 0])
                self[item + "_sigma"] = np.array(self[item][:, :, 1])
                self[item + "_equil"] = np.array(self[item][:, :, 2])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
                self[item + "_grate"] = np.array(self[item][:, :, 3])
                self[item + "_theta"] = np.array(self[item][:, :, 4])
                self[item + "_zeta"] = np.array(self[item][:, :, 5])
                self[item + "_k"] = np.array(self[item][:, :, 6])
            elif item == "B_PROBES":
                self[item + "_target"] = np.array(self[item][:, :, 4])
                self[item + "_sigma"] = np.array(self[item][:, :, 5])
                self[item + "_equil"] = np.array(self[item][:, :, 6])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
                self[item + "_X"] = np.array(self[item][:, :, 0])
                self[item + "_Y"] = np.array(self[item][:, :, 1])
                self[item + "_Z"] = np.array(self[item][:, :, 2])
                self[item + "_MODB"] = np.array(self[item][:, :, 3])
            elif item in ["FLUXLOOPS", "SEGROG"]:
                self[item + "_target"] = np.array(self[item][:, :, 0])
                self[item + "_sigma"] = np.array(self[item][:, :, 1])
                self[item + "_equil"] = np.array(self[item][:, :, 2])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
            elif item == "EXTCUR":
                self[item + "_target"] = np.array(self[item][:, :, 0])
                self[item + "_sigma"] = np.array(self[item][:, :, 1])
                self[item + "_equil"] = np.array(self[item][:, :, 2])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
                self[item + "_dex"] = np.array(self[item][:, :, 3])
            elif item in ["SEPARATRIX", "LIMITER"]:
                self[item + "_target"] = np.array(self[item][:, :, 0])
                self[item + "_sigma"] = np.array(self[item][:, :, 1])
                self[item + "_equil"] = np.array(self[item][:, :, 2])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
                self[item + "_R"] = np.array(self[item][:, :, 4])
                self[item + "_PHI"] = np.array(self[item][:, :, 5])
                self[item + "_Z"] = np.array(self[item][:, :, 6])
            elif item in ["TI", "TE", "IOTA", "VPHI", "PRESS", "NE"]:
                self[item + "_target"] = np.array(self[item][:, :, 4])
                self[item + "_sigma"] = np.array(self[item][:, :, 5])
                self[item + "_equil"] = np.array(self[item][:, :, 6])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
                self[item + "_R"] = np.array(self[item][:, :, 0])
                self[item + "_PHI"] = np.array(self[item][:, :, 1])
                self[item + "_Z"] = np.array(self[item][:, :, 2])
                self[item + "_S"] = np.array(self[item][:, :, 3])
            elif item in ["NELINE", "FARADAY", "SXR"]:
                self[item + "_target"] = np.array(self[item][:, :, 0])
                self[item + "_sigma"] = np.array(self[item][:, :, 1])
                self[item + "_equil"] = np.array(self[item][:, :, 2])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
                self[item + "_R0"] = np.array(self[item][:, :, 3])
                self[item + "_PHI0"] = np.array(self[item][:, :, 4])
                self[item + "_Z0"] = np.array(self[item][:, :, 5])
                self[item + "_R1"] = np.array(self[item][:, :, 6])
                self[item + "_PHI1"] = np.array(self[item][:, :, 7])
                self[item + "_Z1"] = np.array(self[item][:, :, 8])
            elif item == "MSE":
                self[item + "_target"] = np.array(self[item][:, :, 4])
                self[item + "_sigma"] = np.array(self[item][:, :, 5])
                self[item + "_equil"] = np.array(self[item][:, :, 8])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
                self[item + "_R"] = np.array(self[item][:, :, 0])
                self[item + "_PHI"] = np.array(self[item][:, :, 1])
                self[item + "_Z"] = np.array(self[item][:, :, 2])
                self[item + "_S"] = np.array(self[item][:, :, 3])
                self[item + "_ER"] = np.array(self[item][:, :, 6])
                self[item + "_EZ"] = np.array(self[item][:, :, 7])
            elif item == "BOOTSTRAP":
                self[item + "_target"] = np.array(self[item][:, :, 0])
                self[item + "_sigma"] = np.array(self[item][:, :, 1])
                self[item + "_equil"] = np.array(self[item][:, :, 2])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
                self[item + "_S"] = np.array(self[item][:, :, 3])
                self[item + "_avg_jdotb"] = np.array(self[item][:, :, 4])
                self[item + "_beam_jdotb"] = np.array(self[item][:, :, 5])
                self[item + "_boot_jdotb"] = np.array(self[item][:, :, 6])
                self[item + "_jBbs"] = np.array(self[item][:, :, 7])
                self[item + "_facnu"] = np.array(self[item][:, :, 8])
                self[item + "_bsnorm"] = np.array(self[item][:, :, 9])
            elif item == "HELICITY":
                self[item + "_target"] = np.array(self[item][:, :, 0])
                self[item + "_sigma"] = np.array(self[item][:, :, 1])
                self[item + "_equil"] = np.array(self[item][:, :, 2])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
                self[item + "_bnorm"] = np.array(self[item][:, :, 3])
            elif item == "HELICITY_FULL":
                self[item + "_target"] = np.array(self[item][:, :, 0])
                self[item + "_sigma"] = np.array(self[item][:, :, 1])
                self[item + "_equil"] = np.array(self[item][:, :, 2])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
                self[item + "_bnorm"] = np.array(self[item][:, :, 3])
                self[item + "_k"] = np.array(self[item][:, :, 4])
                self[item + "_m"] = np.array(self[item][:, :, 5])
                self[item + "_n"] = np.array(self[item][:, :, 6])
            elif item == "TXPORT":
                self[item + "_target"] = np.array(self[item][:, :, 0])
                self[item + "_sigma"] = np.array(self[item][:, :, 1])
                self[item + "_equil"] = np.array(self[item][:, :, 2])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
                self[item + "_S"] = np.array(self[item][:, :, 3])
            elif item == "COIL_BNORM":
                self[item + "_target"] = np.array(self[item][:, :, 0])
                self[item + "_sigma"] = np.array(self[item][:, :, 1])
                self[item + "_equil"] = np.array(self[item][:, :, 2])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
                self[item + "_U"] = np.array(self[item][:, :, 3])
                self[item + "_V"] = np.array(self[item][:, :, 4])
                self[item + "_BNEQ"] = np.array(self[item][:, :, 5])
                self[item + "_BNF"] = np.array(self[item][:, :, 6])
            elif item == "ORBIT":
                self[item + "_target"] = np.array(self[item][:, :, 0])
                self[item + "_sigma"] = np.array(self[item][:, :, 1])
                self[item + "_equil"] = np.array(self[item][:, :, 2])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
                self[item + "_S"] = np.array(self[item][:, :, 3])
            elif item == "J_STAR":
                self[item + "_target"] = np.array(self[item][:, :, 0])
                self[item + "_sigma"] = np.array(self[item][:, :, 1])
                self[item + "_equil"] = np.array(self[item][:, :, 2])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
                self[item + "_AVGJSTAR"] = np.array(self[item][:, :, 3])
                self[item + "_TRAPSJSTAR"] = np.array(self[item][:, :, 4])
                self[item + "_UJSTAR"] = np.array(self[item][:, :, 5])
                self[item + "_k"] = np.array(self[item][:, :, 6])
                self[item + "_IJSTAR"] = np.array(self[item][:, :, 7])
            elif item == "NEO":
                self[item + "_target"] = np.array(self[item][:, :, 0])
                self[item + "_sigma"] = np.array(self[item][:, :, 1])
                self[item + "_equil"] = np.array(self[item][:, :, 2])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
                self[item + "_k"] = np.array(self[item][:, :, 3])
            elif item == "JDOTB":
                self[item + "_target"] = np.array(self[item][:, :, 0])
                self[item + "_sigma"] = np.array(self[item][:, :, 1])
                self[item + "_equil"] = np.array(self[item][:, :, 2])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
                self[item + "_S"] = np.array(self[item][:, :, 3])
            elif item == "JTOR":
                self[item + "_target"] = np.array(self[item][:, :, 0])
                self[item + "_sigma"] = np.array(self[item][:, :, 1])
                self[item + "_equil"] = np.array(self[item][:, :, 2])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
                self[item + "_S"] = np.array(self[item][:, :, 3])
            elif item == "DKES":
                self[item + "_target"] = np.array(self[item][:, :, 0])
                self[item + "_sigma"] = np.array(self[item][:, :, 1])
                self[item + "_equil"] = np.array(self[item][:, :, 2])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
                self[item + "_S"] = np.array(self[item][:, :, 3])
                self[item + "_NU"] = np.array(self[item][:, :, 4])
                self[item + "_ER"] = np.array(self[item][:, :, 5])
                self[item + "_L11P"] = np.array(self[item][:, :, 6])
                self[item + "_L11M"] = np.array(self[item][:, :, 7])
                self[item + "_L33P"] = np.array(self[item][:, :, 8])
                self[item + "_L33M"] = np.array(self[item][:, :, 9])
                self[item + "_L31P"] = np.array(self[item][:, :, 10])
                self[item + "_L31M"] = np.array(self[item][:, :, 11])
                self[item + "_SCAL11"] = np.array(self[item][:, :, 12])
                self[item + "_SCAL33"] = np.array(self[item][:, :, 13])
                self[item + "_SCAL31"] = np.array(self[item][:, :, 14])
            elif item == "GAMMA_C":
                self[item + "_target"] = np.array(self[item][:, :, 0])
                self[item + "_sigma"] = np.array(self[item][:, :, 1])
                self[item + "_equil"] = np.array(self[item][:, :, 2])
                self[item + "_K"] = np.array(self[item][:, :, 3])
                self[item + "_chisq"] = (
                    (self[item + "_target"] - self[item + "_equil"])
                    / self[item + "_sigma"]
                ) ** 2
        self["chisq"] = ((self["TARGETS"] - self["VALS"]) / self["SIGMAS"]) ** 2
        return

    def plot(self, ax=None, all=True, **kwargs):
        import itertools

        marker = itertools.cycle(("s", "+", "^", "o", "*"))
        # plot the overall iterations
        # set default plotting parameters
        if kwargs.get("linewidth") == None:
            kwargs.update({"linewidth": 2.0})  # prefer thicker lines
        if kwargs.get("label") == None:
            kwargs.update({"label": "Total_Chisq"})  # default label
        f, ax = get_figure(ax)
        ax.semilogy(self["ITER"], np.sum(self["chisq"], axis=1), **kwargs)
        if all:
            for key in self.keys():
                if "_chisq" in key:
                    label = key.replace("_chisq", "")
                    ax.semilogy(
                        self["ITER"],
                        np.sum(self[key], axis=1),
                        label=label,
                        marker=next(marker),
                        linestyle="",
                    )
            plt.legend()
        plt.xlabel("iterations")
        plt.ylabel("chisq")
        return ax

    def plot_helicity(
        self,
        it=-1,
        ordering=0,
        mn=(None, None),
        ax=None,
        log=True,
        normalize=False,
        logical_not=False,
        **kwargs
    ):
        """Plot |B| components in Boozer coordinates from BOOZ_XFORM

        Args:
            it (int, optional): Iteration index to be plotted. Defaults to -1.
            ordering (integer, optional): Plot the leading Nordering asymmetric modes. Defaults to 0.
            mn (tuple, optional): Plot the particular (m,n) mode. Defaults to (None, None).
            ax (Matplotlib axis, optional): Matplotlib axis to be plotted on. Defaults to None.
            log (bool, optional): Plot in log scale. Defaults to True.
            normalize(bool, optionl): Normalized to B_00. Defaults to False.
            logical_not (bool, optional): Unselect mn modes. Defaults to False.
            kwargs (dict): Keyword arguments for matplotlib.pyplot.plot. Defaults to {}.

        Returns:
            ax (Matplotlib axis): Matplotlib axis plotted on.
        """
        from .booz_xform import BOOZ_XFORM

        xs = self["HELICITY_FULL_k"]
        xs = np.array(np.unique(xs), dtype=int)
        ns = len(xs)
        xx = (xs - 1) / np.max(xs)  # max(xs) might be different from NS
        vals = np.reshape(self["HELICITY_FULL_equil"][it], (ns, -1))
        xm = np.reshape(np.array(self["HELICITY_FULL_m"][it], dtype=int), (ns, -1))
        xn = np.reshape(np.array(self["HELICITY_FULL_n"][it], dtype=int), (ns, -1))
        return BOOZ_XFORM.plot_helicity(
            vals,
            xm[0, :],
            xn[0, :],
            xx,
            ordering,
            mn,
            ax,
            log,
            normalize,
            logical_not,
            **kwargs
        )

    def plot_balloon(self, it=-1, ax=None, **kwargs):
        """Plot Ballooning instability from COBRAVMEC

        Args:
            it (int, optional): Iteration index to be plotted. Defaults to -1.
            ax (Matplotlib axis, optional): Matplotlib axis to be plotted on. Defaults to None.
            kwargs (dict): Keyword arguments for matplotlib.pyplot.plot. Defaults to {}.

        Returns:
            ax (Matplotlib axis): Matplotlib axis plotted on.
        """
        fig, ax = get_figure(ax)
        ax.plot(self["BALLOON_k"][it], self["BALLOON_grate"][it], "o")
        ax.set_xlabel("Radial Grid")
        ax.set_ylabel("Growth Rate")
        ax.set_title("COBRA Ballooning Stability (<0 Stable)")
        return ax

    def plot_neo(self, it=-1, ax=None, **kwargs):
        """Plot effective ripple from NEO

        Args:
            it (int, optional): Iteration index to be plotted. Defaults to -1.
            ax (Matplotlib axis, optional): Matplotlib axis to be plotted on. Defaults to None.
            kwargs (dict): Keyword arguments for matplotlib.pyplot.plot. Defaults to {}.

        Returns:
            ax (Matplotlib axis): Matplotlib axis plotted on.
        """
        fig, ax = get_figure(ax)
        xs = self["NEO_k"][it]
        xx = (xs - 1) / np.max(xs)  # max(xs) might be different from NS
        ax.semilogy(xx, self["NEO_equil"][it], "o")
        ax.set_xlabel("Normalized flux (s) ")
        ax.set_ylabel(r"${\epsilon_{\mathrm{eff}}}^{3/2}$")
        ax.set_title("Effective ripple from NEO")
        return ax



import numpy as np
import matplotlib.pyplot as plt
from .sortedDict import SortedDict
from .misc import trig2real

__all__ = ["BOOZ_XFORM"]


class BOOZ_XFORM(SortedDict):
    def __init__(self, filename, **kwargs):
        import xarray

        SortedDict.__init__(self)
        self.output = xarray.open_dataset(filename)
        self.ns = int(self.output["ns_b"].values)
        self.nfp = int(self.output["nfp_b"].values)
        self.xm = self.output["ixm_b"].values
        self.xn = self.output["ixn_b"].values / self.nfp
        self.jlist = self.output["jlist"].values
        self.bmnc = self.output["bmnc_b"].values
        return

    @staticmethod
    def write_input(extension, mbooz, nbooz, surfaces):
        """Write a BOO_XFORM input file

        Args:
            extension (str): VMEC output extension.
            mbooz (int): Maximum poloidal mode.
            nbooz (int): Maximum toroidal mode
            surfaces (list): List of flux surfaces.
        """
        template = """{mbooz:d}  {nbooz:d}
'{extension:s}'
{surfaces:s}
        """
        assert mbooz > 0, "mbooz should be >0."
        assert nbooz > 0, "nbooz should be >0."
        surfaces = ["{:} ".format(i) for i in surfaces]
        surfaces = "".join(surfaces)
        with open("inbooz." + extension, "w") as f:
            f.write(
                template.format(
                    mbooz=mbooz, nbooz=nbooz, extension=extension, surfaces=surfaces
                )
            )
        return

    @staticmethod
    def from_vmec(wout, mbooz=0, nbooz=0, surfaces=[]):
        """Prepare BOOZ_XFORM input file from VMEC wout file

        Args:
            wout (str): VMEC wout filename.
            mbooz (int, optional): Maximum poloildal mode number. Defaults to 0 (4*Mpol).
            nbooz (int, optional): Maximum toroidal mode number. Defaults to 0 (4*Ntor).
            surfaces (list, optional): Flux surfaces list. Defaults to [] (1:NS).

        """
        import xarray

        vmec = xarray.open_dataset(wout)
        ind = wout.index("wout_") + 5
        end = wout.index(".nc")
        extension = wout[ind:end]
        if mbooz == 0:
            mbooz = 4 * int(vmec["mpol"].values) + 1
        if nbooz == 0:
            nbooz = 4 * int(vmec["ntor"].values)
        if len(surfaces) == 0:
            ns = int(vmec["ns"].values)
            surfaces = 1 + np.arange(ns)
        vmec.close()
        return BOOZ_XFORM.write_input(extension, mbooz, nbooz, surfaces)

    def plot(
        self,
        ordering=0,
        mn=(None, None),
        ax=None,
        log=True,
        normalize=True,
        logical_not=False,
        **kwargs
    ):
        """Plot |B| components 1D profile

        Args:
            ordering (integer, optional): Plot the leading Nordering asymmetric modes. Defaults to 0.
            mn (tuple, optional): Plot the particular (m,n) mode. Defaults to (None, None).
            ax (Matplotlib axis, optional): Matplotlib axis to be plotted on. Defaults to None.
            log (bool, optional): Plot in log scale. Defaults to True.
            normalize(bool, optionl): Normalized to B_00. Defaults to True.
            logical_not (bool, optional): Unselect mn modes. Defaults to False.
            kwargs (dict): Keyword arguments for matplotlib.pyplot.plot. Defaults to {}.

        Returns:
            ax (Matplotlib axis): Matplotlib axis plotted on.
        """
        xx = self.jlist / self.ns
        return self.plot_helicity(
            self.bmnc,
            self.xm,
            self.xn,
            xx,
            ordering,
            mn,
            ax,
            log,
            normalize,
            logical_not,
            **kwargs
        )

    @staticmethod
    def plot_helicity(
        vals,
        xm,
        xn,
        xx,
        ordering=0,
        mn=(None, None),
        ax=None,
        log=False,
        normalize=False,
        logical_not=False,
        **kwargs
    ):
        # get figure and ax data
        if ax is None:
            fig, ax = plt.subplots()
        plt.sca(ax)
        # check if normalizing to B_00
        if normalize:
            try:
                vals /= vals[:, np.logical_and(xm == 0, xn == 0)]
            except ValueError:
                print("Something wrong with the normalization to B_00")
        # determine filter condition
        if mn[0] is not None:
            mfilter = xm == mn[0]
            m = "m={:}".format(mn[0])
        else:
            mfilter = np.full(np.shape(xm), True)
            m = "m"
        if mn[1] is not None:
            nfilter = xn == mn[1]
            n = "n={:}".format(mn[1])
        else:
            nfilter = np.full(np.shape(xn), True)
            n = "n"
        cond = np.logical_and(mfilter, nfilter)
        if logical_not:
            cond = np.logical_not(cond)
        # data = np.reshape(vals[:, cond], (ns, -1))
        vals = vals[:, cond]
        xm = xm[cond]
        xn = xn[cond]
        # select the top ordering asymmetric terms
        if ordering:
            assert ordering >= 1
            data = np.linalg.norm(vals, axis=0)
            ind_arg = np.argsort(data)
            for i in range(ordering):
                ind = ind_arg[-1 - i]  # index of the i-th largest term
                m = xm[ind]
                n = xn[ind]
                if m == 0 and n == 0:
                    continue
                kwargs["label"] = "m={:}, n={:}".format(int(m), int(n))
                if log:
                    ax.semilogy(xx, np.abs(vals[:, ind]), **kwargs)
                else:
                    ax.plot(xx, np.abs(vals[:, ind]), **kwargs)
            ylabel = r"$\frac{B_{m,n}}{ B_{00} }$"
        else:
            if log:
                line = ax.semilogy(xx, np.linalg.norm(vals, axis=1), **kwargs)
            else:
                line = ax.plot(xx, np.linalg.norm(vals, axis=1), **kwargs)
            ylabel = r"$ \frac{{ \Vert B_{{ {:},{:} }} \Vert }}{{ B_{{00}} }} $".format(
                m, n
            )
        plt.xlabel("normalized flux (s)", fontsize=16)
        plt.ylabel(ylabel, fontsize=16)
        plt.xticks(fontsize=15)
        plt.yticks(fontsize=15)
        return ax

    def plot2d(self, ns=-1, npol=128, ntor=128, ax=None, contour=True, **kwargs):
        """Plot |B| contour on the flux surface

        Args:
            ns (int, optional): Flux surface index in self.jlist. Defaults to -1.
            npol (int, optional): Poloidal resolution. Defaults to 128.
            ntor (int, optional): Toroidal resolution. Defaults to 128.
            ax (Matplotlib axis, optional): Matplotlib axis to be plotted on. Defaults to None.
            contour (bool, optional): Whether plotting the contour lines. Defaults to True.

        Returns:
            ax (Matplotlib axis): Matplotlib axis plotted on.
        """
        _theta = np.linspace(0, np.pi * 2, npol, endpoint=True)
        _zeta = np.linspace(0, np.pi * 2, ntor, endpoint=True)
        modB = trig2real(_theta, _zeta, self.xm, self.xn, self.bmnc[ns, :])
        # get figure and ax data
        if ax is None:
            fig, ax = plt.subplots()
        plt.sca(ax)
        bl = ax.imshow(modB, origin="lower", extent=[0, 2 * np.pi, 0, 2 * np.pi])
        plt.xticks(
            np.linspace(0, 2 * np.pi, 3),
            ["0", r"$\pi/{:d}$".format(self.nfp), r"$2\pi/{:d}$".format(self.nfp)],
            fontsize=15,
        )
        plt.yticks(
            np.linspace(0, 2 * np.pi, 3), ["0", r"$\pi$", r"$2\pi$"], fontsize=15
        )
        if contour:
            ax.contour(_zeta, _theta, modB, colors="grey")
        clb = plt.colorbar(bl)
        clb.ax.tick_params(labelsize=15)
        clb.ax.set_title("|B|(T)", fontsize=15)
        plt.xlabel(r"$\zeta_{Boozer}$", fontsize=16)
        plt.ylabel(r"$\theta_{Boozer}$", fontsize=16)
        import matplotlib.ticker as tck

        ax.xaxis.set_minor_locator(tck.AutoMinorLocator())
        ax.yaxis.set_minor_locator(tck.AutoMinorLocator())
        return ax

    def to_FOCUS(self, ns=-1, focus_file="plasma.boundary", tol=1.0e-8, iota=None):
        """Write a FOCUS plasma boundary file in Boozer coordinates

        Args:
            ns (int, optional): Flux surface label to be exported. Defaults to -1.
            focus_file (str, optional): FOCUS plasma boundary file name. Defaults to 'plasma.boundary'.
            tol (float, optional): Truncated tolerance. Defaults to 1.0E-8.
            iota (float, optional): The target iota to be extracted. Defaults to None.
        """
        mn = int(self.output["mnboz_b"].values)
        # find the closest ns if specifying an iota
        if iota is not None:
            ns = np.argmin(np.abs(self.output["iota_b"].values - iota))
            print(
                "Target iota: ",
                iota,
                ", closest value: ",
                self.output["iota_b"].values[ns],
            )
        rbc = np.array(self.output["rmnc_b"][ns, :])
        # rbs = np.zeros(mn)
        zbs = np.array(self.output["zmns_b"][ns, :])
        # zbc = np.zeros(mn)
        pmns = np.array(self.output["pmns_b"][ns, :])
        # pmnc = np.zeros(mn)
        Nbnf = 0
        # count non-zero terms
        amn = 0
        for imn in range(mn):
            if (abs(rbc[imn]) + abs(zbs[imn] + abs(pmns[imn]))) > tol:
                amn += 1  # number of nonzero coef.
        # write FOCUS plasma boundary file
        with open(focus_file, "w") as fofile:
            fofile.write("# bmn   bNfp   nbf " + "\n")
            fofile.write("{:d} \t {:d} \t {:d} \n".format(amn, self.nfp, Nbnf))
            fofile.write("#plasma boundary" + "\n")
            fofile.write("# n m Rbc Rbs Zbc Zbs Pmnc Pmns" + "\n")
            for imn in range(mn):
                if (abs(rbc[imn]) + abs(zbs[imn] + abs(pmns[imn]))) > tol:
                    fofile.write(
                        "{:4d}  {:4d} \t {:23.15E}  {:12.5E}  {:12.5E} {:23.15E} {:12.5E}  {:23.15E} \n".format(
                            int(self.xn[imn]),
                            self.xm[imn],
                            rbc[imn],
                            0.0,
                            0.0,
                            zbs[imn],
                            0.0,
                            pmns[imn],
                        )
                    )
            fofile.write("#Bn harmonics \n")
            fofile.write("# n m bnc bns" + "\n")
        return



import numpy as np

u0_d_4pi = 1.0e-7


class SingleCoil(object):
    """Python class representing a single coil as discrete points in Cartesian coordinates.

    Attributes
    ----------
    x: Data in x-coordinate

    Args:
        x (list, optional): Data in x-coordinate. Defaults to [].
        y (list, optional): Data in y-coordinate. Defaults to [].
        z (list, optional): Data in z-coordinate Defaults to [].
        I (float, optional): Coil current. Defaults to 0.0.
        name (str, optional): Coil name. Defaults to "coil1".
        group (int, optional): Coil group for labeling. Defaults to 1.
    """

    def __init__(self, x=[], y=[], z=[], I=0.0, name="coil1", group=1):
        assert len(x) == len(y) == len(z), "dimension not consistent"
        self.x = np.asarray(x)
        self.y = np.asarray(y)
        self.z = np.asarray(z)
        self.I = I
        self.name = name
        self.group = group
        self.xt = None
        self.yt = None
        self.zt = None
        return

    def bfield(self, pos):
        """Calculate the magnetic field at an arbitrary point using `self.dt`.

        Args:
            pos (list): Evaluation point in Cartesian coordinates.

        Returns:
            numpy.ndarray: B vector produced by the coil.
        """
        ob_pos = np.atleast_1d(pos)
        dx = ob_pos[0] - self.x[:-1]
        dy = ob_pos[1] - self.y[:-1]
        dz = ob_pos[2] - self.z[:-1]
        dr = dx * dx + dy * dy + dz * dz
        Bx = (dz * self.yt[:-1] - dy * self.zt[:-1]) * np.power(dr, -1.5) * self.dt
        By = (dx * self.zt[:-1] - dz * self.xt[:-1]) * np.power(dr, -1.5) * self.dt
        Bz = (dy * self.xt[:-1] - dx * self.yt[:-1]) * np.power(dr, -1.5) * self.dt
        B = np.array([np.sum(Bx), np.sum(By), np.sum(Bz)]) * u0_d_4pi * self.I
        return B

    def bfield_fd(self, pos):
        """Calculate the magnetic field at an arbitrary point using finite difference.

        Args:
            pos (list): Evaluation point in Cartesian coordinates.

        Returns:
            numpy.ndarray: B vector produced by the coil.
        """
        pos = np.atleast_1d(pos)
        xt = self.x[1:] - self.x[:-1]
        yt = self.y[1:] - self.y[:-1]
        zt = self.z[1:] - self.z[:-1]
        dx = pos[0] - (self.x[:-1] + self.x[1:]) / 2
        dy = pos[1] - (self.y[:-1] + self.y[1:]) / 2
        dz = pos[2] - (self.z[:-1] + self.z[1:]) / 2
        dr = dx * dx + dy * dy + dz * dz
        Bx = (dz * yt - dy * zt) * np.power(dr, -1.5)
        By = (dx * zt - dz * xt) * np.power(dr, -1.5)
        Bz = (dy * xt - dx * yt) * np.power(dr, -1.5)
        B = np.array([np.sum(Bx), np.sum(By), np.sum(Bz)]) * u0_d_4pi * self.I
        return B

    def bfield_HH(self, pos, **kwargs):
        """Calculate B field at an arbitrary point using the Hanson-Hirshman expression

        Arguments:
            pos (list): Cartesian coordinates for the evaluation point.

        Returns:
            numpy.ndarray: B vector produced by the coil.
        """
        xyz = np.array([self.x, self.y, self.z]).T
        pos = np.atleast_2d(pos)
        assert (pos.shape)[1] == 3
        Rvec = pos[:, np.newaxis, :] - xyz[np.newaxis, :, :]
        assert (Rvec.shape)[-1] == 3
        RR = np.linalg.norm(Rvec, axis=2)
        Riv = Rvec[:, :-1, :]
        Rfv = Rvec[:, 1:, :]
        Ri = RR[:, :-1]
        Rf = RR[:, 1:]
        B = (
            np.sum(
                np.cross(Riv, Rfv)
                * ((Ri + Rf) / ((Ri * Rf) * (Ri * Rf + np.sum(Riv * Rfv, axis=2))))[
                    :, :, np.newaxis
                ],
                axis=1,
            )
            * u0_d_4pi
            * self.I
        )
        return B

    def hanson_hirshman(self, pos):
        """Wrapper for the fortran code biotsavart.hanson_hirshman

        Args:
            pos (ndarray, (n,3)): Evaluation points in space

        Returns:
            ndarray, (n,3): Magnetic field at the evaluation point
        """
        from coilpy_fortran import hanson_hirshman

        xyz = np.transpose([self.x, self.y, self.z])
        return hanson_hirshman(pos, xyz, self.I)

    def biot_savart(self, pos):
        from coilpy_fortran import biot_savart

        xyz = np.transpose([self.x, self.y, self.z])
        dxyz = np.transpose([self.xt * self.dt, self.yt * self.dt, self.zt * self.dt])
        return biot_savart(pos, xyz[:-1, :], self.I, dxyz[:-1, :])

    def fourier_tangent(self):
        """
        Approximate the tangent using Fourier representation.
        """
        from .misc import fft_deriv

        if not np.isclose(self.x[0], self.x[-1]):
            print("Warning: Spectral derivatives using FFT are used for closed coils.")
        self.dt = 2 * np.pi / (len(self.x) - 1)
        fftxy = fft_deriv(self.x[:-1] + 1j * self.y[:-1])
        fftz = fft_deriv(self.z[:-1])
        self.xt = np.real(fftxy)
        self.yt = np.imag(fftxy)
        self.zt = np.real(fftz)
        self.xt = np.concatenate((self.xt, self.xt[0:1]))
        self.yt = np.concatenate((self.yt, self.yt[0:1]))
        self.zt = np.concatenate((self.zt, self.zt[0:1]))
        return

    def interpolate(self, num=256, kind="fft", nf=-1):
        """Interpolate to get more data points.

        Args:
            num (int, optional): The total number of points after interpolation. Defaults to 256.
            kind (str, optional): Specifies the kind of interpolation, could be 'fft'
                                  or scipy.interp1d.kind.  Defaults to 'cubic'.
            nf (int, optional): Number of truncated Fourier modes. Defaults to -1.
        """
        from scipy.interpolate import interp1d
        from coilpy.misc import trigfft, trig2real

        cur_len = len(self.x)
        assert cur_len > 0
        theta = np.linspace(0, 1, num=cur_len, endpoint=True)
        theta_new = np.linspace(0, 1, num=num, endpoint=True)
        if kind == "fft":
            # FFT
            fftxy = trigfft(self.x[:-1] + 1j * self.y[:-1], tr=nf)
            fftz = trigfft(self.z[:-1], tr=nf)
            xm = fftxy["n"]
            xc = fftxy["rcos"]
            xs = fftxy["rsin"]
            yc = fftxy["icos"]
            ys = fftxy["isin"]
            zc = fftz["rcos"]
            zs = fftz["rsin"]
            self.x = trig2real(theta_new, zeta=None, xm=xm, xn=None, fmnc=xc, fmns=xs)
            self.y = trig2real(theta_new, zeta=None, xm=xm, xn=None, fmnc=yc, fmns=ys)
            self.z = trig2real(theta_new, zeta=None, xm=xm, xn=None, fmnc=zc, fmns=zs)
        else:
            # splines
            f = interp1d(theta, self.x, kind="cubic")
            self.x = f(theta_new)
            f = interp1d(theta, self.y, kind="cubic")
            self.y = f(theta_new)
            f = interp1d(theta, self.z, kind="cubic")
            self.z = f(theta_new)
        return

    def magnify(self, ratio):
        """Magnify the coil with a ratio.

        Args:
            ratio (float): The magnifying ratio.
        """
        # number of points
        nseg = len(self.x)
        # assuming closed curve; should be revised
        if True:  # abs(self.x[0] - self.x[-1]) < 1.0E-8:
            nseg -= 1
        assert nseg > 1
        # get centroid
        centroid = np.array(
            [
                np.sum(self.x[0:nseg]) / nseg,
                np.sum(self.y[0:nseg]) / nseg,
                np.sum(self.z[0:nseg]) / nseg,
            ]
        )
        # magnify
        for i in range(nseg):
            xyz = np.array([self.x[i], self.y[i], self.z[i]])
            dr = xyz - centroid
            [self.x[i], self.y[i], self.z[i]] = centroid + ratio * dr
        try:
            self.x[nseg] = self.x[0]
            self.y[nseg] = self.y[0]
            self.z[nseg] = self.z[0]
            return
        except ValueError:
            return

    def plot(self, engine="mayavi", fig=None, ax=None, show=True, **kwargs):
        """Plot the coil in a specified engine.

        Args:
            engine (str, optional): Plot enginer, could be {pyplot, mayavi, plotly}. Defaults to "mayavi".
            fig (, optional): Figure to be plotted on. Defaults to None.
            ax (matplotlib.axis, optional): Axis to be plotted on. Defaults to None.
            show (bool, optional): If show the plotly figure immediately. Defaults to True.

        Raises:
            ValueError: Invalid engine option, should be one of {pyplot, mayavi, plotly}.
        """
        if engine == "pyplot":
            import matplotlib.pyplot as plt

            # plot in matplotlib.pyplot
            if ax is None or ax.name != "3d":
                fig = plt.figure()
                ax = fig.add_subplot(111, projection="3d")
            ax.plot(self.x, self.y, self.z, **kwargs)
        elif engine == "mayavi":
            # plot 3D line in mayavi.mlab
            from mayavi import mlab  # to overrid plt.mlab

            mlab.plot3d(self.x, self.y, self.z, **kwargs)
        elif engine == "plotly":
            import plotly.graph_objects as go

            if "color" in list(kwargs.keys()):
                color = kwargs["color"]
                del kwargs["color"]
            else:
                color = "blue"
            kwargs.setdefault("line", go.scatter3d.Line(color=color, width=4))
            if fig is None:
                fig = go.Figure()
            fig.add_trace(
                go.Scatter3d(
                    x=self.x, y=self.y, z=self.z, mode="lines", name=self.name, **kwargs
                )
            )
            fig.update_layout(scene_aspectmode="data")
            if show:
                fig.show()
        else:
            raise ValueError("Invalid engine option {pyplot, mayavi, plotly}")
        return fig

    def plot2d(
        self,
        engine="mayavi",
        fig=None,
        ax=None,
        show=True,
        width=0.1,
        height=0.1,
        frame="centroid",
        **kwargs
    ):
        """Plot the coil with finite size.

        Args:
            engine (str, optional): Plot enginer, could be {pyplot, mayavi, plotly}. Defaults to "mayavi".
            fig (, optional): Figure to be plotted on. Defaults to None.
            ax (matplotlib.axis, optional): Axis to be plotted on. Defaults to None.
            show (bool, optional): If show the plotly figure immediately. Defaults to True.
            width (float, optional): Coil width. Defaults to 0.1.
            height (float, optional): Coil height. Defaults to 0.1.
            frame (str, optional): Finite-build frame, could be one of
                                  ("centroid", "frenet", "parallel"). Defaults to "centroid".
        """
        xx, yy, zz = self.rectangle(width, height, frame)
        if engine == "pyplot":
            pass
        elif engine == "mayavi":
            # plot 3D line in mayavi.mlab
            from mayavi import mlab  # to overrid plt.mlab

            mlab.mesh(xx, yy, zz, **kwargs)
        elif engine == "plotly":
            import plotly.graph_objects as go

            if "color" in list(kwargs.keys()):
                color = kwargs["color"]
                del kwargs["color"]
                kwargs["colorscale"] = [[0, color], [1, color]]
            kwargs.setdefault("showscale", False)
            if fig is None:
                fig = go.Figure()
            fig.add_trace(go.Surface(x=xx, y=yy, z=zz, **kwargs))
            fig.update_layout(scene_aspectmode="data")
            if show:
                fig.show()
        else:
            raise ValueError("Invalid engine option {pyplot, mayavi, plotly}")
        return

    def rectangle(self, width=0.1, height=0.1, frame="centroid", **kwargs):
        """Expand single coil filament to a finite-build coil.

        Args:
            width (float, optional): Coil width. Defaults to 0.1.
            height (float, optional): Coil height. Defaults to 0.1.
            frame (str, optional): Finite-build frame, could be one of
                                  ("centroid", "frenet", "parallel"). Defaults to "centroid".

        Returns:
            numpy.ndarry: x-coordiante for plotting as a mesh.
            numpy.ndarry: y-coordiante for plotting as a mesh.
            numpy.ndarry: z-coordiante for plotting as a mesh.
        """
        n = np.size(self.x)
        # calculate the tangent
        if self.xt is None:
            self.spline_tangent()
        xt = self.xt
        yt = self.yt
        zt = self.zt
        # dt = 2 * np.pi / (n - 1)
        # xt = np.gradient(self.x)/dt
        # yt = np.gradient(self.y)/dt
        # zt = np.gradient(self.z)/dt
        tt = np.sqrt(xt * xt + yt * yt + zt * zt)
        xt = xt / tt
        yt = yt / tt
        zt = zt / tt

        # use surface normal if needed
        if frame == "centroid":
            # use the geometry center is a good idea
            center_x = np.average(self.x[0 : n - 1])
            center_y = np.average(self.y[0 : n - 1])
            center_z = np.average(self.z[0 : n - 1])
            xn = self.x - center_x
            yn = self.y - center_y
            zn = self.z - center_z
            nt = xn * xt + yn * yt + zn * zt
            xn = xn - nt * xt
            yn = yn - nt * yt
            zn = zn - nt * zt
        elif frame == "frenet":
            self.spline_tangent(der=2)
            xn = self.xa
            yn = self.ya
            zn = self.za
        elif frame == "parallel":
            # parallel transport frame
            # Hanson & Ma, Parallel Transp ort Approach to Curve Framing, 1995
            def rotate(x, ang):
                c = np.cos(ang)
                s = np.sin(ang)
                return [
                    [
                        c + x[0] ** 2 * (1 - c),
                        x[0] * x[1] * (1 - c) - s * x[2],
                        x[2] * x[0] * (1 - c) + s * x[1],
                    ],
                    [
                        x[0] * x[1] * (1 - c) + s * x[2],
                        c + x[1] ** 2 * (1 - c),
                        x[2] * x[1] * (1 - c) - s * x[0],
                    ],
                    [
                        x[0] * x[2] * (1 - c) - s * x[1],
                        x[1] * x[2] * (1 - c) + s * x[0],
                        c + x[2] ** 2 * (1 - c),
                    ],
                ]

            T = np.transpose([self.xt, self.yt, self.zt])
            T = T / np.linalg.norm(T, axis=1)[:, np.newaxis]
            B = np.cross(T[:-1], T[1:], axis=1)
            B = B / np.linalg.norm(B, axis=1)[:, np.newaxis]
            theta = np.arccos(np.sum(T[:-1] * T[1:], axis=1))
            V = np.zeros_like(T)
            kwargs.setdefault("vx", self.x[0] - np.average(self.x[0:-1]))
            kwargs.setdefault("vy", self.y[0] - np.average(self.y[0:-1]))
            vx = kwargs["vx"]
            vy = kwargs["vy"]
            vz = -(vx * T[0, 0] + vy * T[0, 1]) / T[0, 2]
            vv = np.linalg.norm([vx, vy, vz])
            V[0, :] = [vx / vv, vy / vv, vz / vv]
            print(np.dot(V[0, :], T[0, :]))
            for i in range(len(theta)):
                V[i + 1, :] = rotate(B[i, :], theta[i]) @ V[i, :]
            xn = V[:, 0]
            yn = V[:, 1]
            zn = V[:, 2]
        else:
            assert True, "not finished"

        nn = np.sqrt(xn * xn + yn * yn + zn * zn)
        xn = xn / nn
        yn = yn / nn
        zn = zn / nn
        # calculate the bi-normal
        xb = yt * zn - yn * zt
        yb = zt * xn - zn * xt
        zb = xt * yn - xn * yt
        bb = np.sqrt(xb * xb + yb * yb + zb * zb)
        xb = xb / bb
        yb = yb / bb
        zb = zb / bb
        # get the boundary lines
        z1 = self.z - width / 2 * zb + height / 2 * zn
        x1 = self.x - width / 2 * xb + height / 2 * xn
        x2 = self.x + width / 2 * xb + height / 2 * xn
        y2 = self.y + width / 2 * yb + height / 2 * yn
        z2 = self.z + width / 2 * zb + height / 2 * zn
        x3 = self.x + width / 2 * xb - height / 2 * xn
        y3 = self.y + width / 2 * yb - height / 2 * yn
        z3 = self.z + width / 2 * zb - height / 2 * zn
        x4 = self.x - width / 2 * xb - height / 2 * xn
        y4 = self.y - width / 2 * yb - height / 2 * yn
        z4 = self.z - width / 2 * zb - height / 2 * zn
        y1 = self.y - width / 2 * yb + height / 2 * yn
        # assemble
        xx = np.array([x1, x2, x3, x4, x1])
        yy = np.array([y1, y2, y3, y4, y1])
        zz = np.array([z1, z2, z3, z4, z1])
        return xx, yy, zz

    def spline_tangent(self, order=3, der=1):
        """Calculate the tangent of coil using spline interpolation

        Args:
            order (int, optional): Order of spline interpolation used. Defaults to 3.
        """
        from scipy import interpolate

        t = np.linspace(0, 2 * np.pi, len(self.x), endpoint=True)
        self.dt = 2 * np.pi / (len(self.x) - 1)
        fx = interpolate.splrep(t, self.x, s=0, k=order)
        fy = interpolate.splrep(t, self.y, s=0, k=order)
        fz = interpolate.splrep(t, self.z, s=0, k=order)
        self.xt = interpolate.splev(t, fx, der=1)
        self.yt = interpolate.splev(t, fy, der=1)
        self.zt = interpolate.splev(t, fz, der=1)
        if der == 2:
            self.xa = interpolate.splev(t, fx, der=2)
            self.ya = interpolate.splev(t, fy, der=2)
            self.za = interpolate.splev(t, fz, der=2)
        return

    def toVTK(self, vtkname, **kwargs):
        """Write the coil as a VTK file

        Args:
            vtkname (string): VTK filename
        """
        from pyevtk.hl import polyLinesToVTK

        kwargs.setdefault("cellData", {})
        kwargs["cellData"].setdefault("I", np.array([self.I]))
        polyLinesToVTK(
            vtkname,
            np.array(self.x),
            np.array(self.y),
            np.array(self.z),
            np.array([len(self.x)]),
            **kwargs
        )
        return


class Coil(object):
    """Python object for a set of coils.

    Args:
        xx (list, optional): Coil data in x-coordinates. Defaults to [[]].
        yy (list, optional): Coil data in y-coordinates. Defaults to [[]].
        zz (list, optional): Coil data in z-coordinates. Defaults to [[]].
        II (list, optional): Coil currents. Defaults to [[]].
        names (list, optional): Coil names. Defaults to [[]].
        groups (list, optional): Coil groups. Defaults to [[]].

    A convenient way for construction is to use `self.read_makegrid(filename)`, like

        ``
        coil = CoilSet.read_makegrid('coils.sth')
        ``

    Each coil is stored in `self.data` in the format of `coilpy.coils.SingleCoil`.

    You can plot the coilset using `self.plot`.

    The coilset can be saved in the format of MAKEGRID using `self.save_makegrid`
    and saved as VTK files using `self.toVTK`.
    """

    def __init__(self, xx=[], yy=[], zz=[], II=[], names=[], groups=[]):
        assert (
            len(xx) == len(yy) == len(zz) == len(II) == len(names) == len(groups)
        ), "dimension not consistent"
        self.num = len(xx)
        self.data = []
        for i in range(self.num):
            self.data.append(
                SingleCoil(
                    x=xx[i], y=yy[i], z=zz[i], I=II[i], name=names[i], group=groups[i]
                )
            )
        self.index = 0
        return

    def __iter__(self):
        return self

    def __next__(self):
        if self.index < self.num:
            self.index += 1
            return self.data[self.index - 1]
        else:
            self.index = 0
            raise StopIteration()

    def __len__(self):
        return self.num

    def __add__(self, other):
        """Join two coil sets.

        Args:
            other (`coilpy.coil.Coil`): The coilset to be added.

        Returns:
            `coilpy.coil.Coil`: The total coil set.
        """
        from copy import deepcopy

        total = deepcopy(self)
        total.num += other.num
        total.data += other.data
        total.index = 0
        return total

    @classmethod
    def read_makegrid(cls, filename):
        """Read coils from the MAKEGRID format.

        Args:
            filename (str): file path and name

        Raises:
            IOError: Check if file exists

        Returns:
            Coil class: the python class Coil
        """
        import os

        # check existence
        if not os.path.exists(filename):
            raise IOError("File not existed. Please check again!")
        # read and parse data
        cls.header = ""
        with open(filename, "r") as coilfile:  # read coil xyz and I
            cls.header = "".join(
                (coilfile.readline(), coilfile.readline(), coilfile.readline())
            )
            icoil = 0
            xx = [[]]
            yy = [[]]
            zz = [[]]
            II = []
            names = []
            groups = []
            tmpI = 0.0
            for line in coilfile:
                linelist = line.split()
                if len(linelist) < 4:
                    # print("End of file or invalid format!")
                    break
                xx[icoil].append(float(linelist[0]))
                yy[icoil].append(float(linelist[1]))
                zz[icoil].append(float(linelist[2]))
                if len(linelist) == 4:
                    tmpI = float(linelist[-1])
                if len(linelist) > 4:
                    II.append(tmpI)
                    try:
                        group = int(linelist[4])
                    except ValueError:
                        group = len(groups) + 1
                    groups.append(group)
                    try:
                        name = linelist[5]
                    except IndexError:
                        name = "coil"
                    names.append(name)
                    icoil = icoil + 1
                    xx.append([])
                    yy.append([])
                    zz.append([])
        xx.pop()
        yy.pop()
        zz.pop()
        # print(len(xx) , len(yy) , len(zz) , len(II) , len(names) , len(groups))
        return cls(xx=xx, yy=yy, zz=zz, II=II, names=names, groups=groups)

    @classmethod
    def read_gpec_coils(cls, filename, current=1.0):
        """Read coils from GPEC files.

        Args:
            filename (str): File name.
            current (float, optional): Coil current. Defaults to 1.0.

        Returns:
            Coil: Coil object.
        """
        import os

        with open(filename, "r") as f:
            line1 = f.readline()
            ncoil, s, nsec, nw = list(map(int, list(map(float, line1.split()))))
        x, y, z = np.genfromtxt(filename, skip_header=1).T.reshape(3, ncoil, s, nsec)
        c = nw
        # not quite sure what is s
        xx = x[:, 0, :]
        yy = y[:, 0, :]
        zz = z[:, 0, :]
        II = nw * np.ones(ncoil) * current
        names = [os.path.split(filename)[-1].split(".")[0] for i in range(ncoil)]
        groups = range(1, ncoil + 1)
        return cls(xx=xx, yy=yy, zz=zz, II=II, names=names, groups=groups)

    def plot(
        self,
        irange=[],
        engine="pyplot",
        plot2d=False,
        ax=None,
        fig=None,
        show=True,
        **kwargs
    ):
        """Plot coils in mayavi or matplotlib or plotly.

        Args:
            irange (list, optional): Coil list to be plotted. Defaults to [].
            engine (string, optional): Plotting engine. One of {'pyplot', 'mayavi', 'plotly'}. Defaults to "pyplot".
            plot2d (bool, optional): If plotting with finite size. Defaults to False.
            fig (, optional): figure to be plotted on. Defaults to None.
            ax (, optional): axis to be plotted on. Defaults to None.
            show (bool, optional): if show the plotly figure immediately. Defaults to True.
            kwargs (dict, optional): Keyword dict for plotting settings.
        """
        if len(irange) == 0:
            irange = range(self.num)
        if engine == "pyplot":
            import matplotlib.pyplot as plt

            # plot in matplotlib.pyplot
            if ax is None or ax.name != "3d":
                fig = plt.figure()
                ax = fig.add_subplot(111, projection="3d")
        elif engine == "plotly":
            import plotly.graph_objects as Go

            if fig is None:
                fig = Go.Figure()
                ax = None
        for i in irange:
            if engine == "plotly":
                kwargs["legendgroup"] = self.data[i].group
                kwargs["show"] = False
            if plot2d:
                self.data[i].plot2d(engine=engine, fig=fig, ax=ax, **kwargs)
            else:
                self.data[i].plot(engine=engine, fig=fig, ax=ax, **kwargs)
        if engine == "plotly":
            if show:
                fig.show()
        return

    def save_makegrid(self, filename, nfp=1, **kwargs):
        """Write coils in the MAKEGRID format.

        Args:
            filename (str): File name and path.
            nfp (int, optional): Number of toroidal periodicity. Defaults to 1.
        """
        assert len(self) > 0
        with open(filename, "w") as wfile:
            wfile.write("periods {:3d} \n".format(nfp))
            wfile.write("begin filament \n")
            wfile.write("mirror NIL \n")
            for icoil in list(self):
                Nseg = len(icoil.x)  # number of segments;
                assert Nseg > 1
                for iseg in range(Nseg - 1):  # the last point match the first one;
                    wfile.write(
                        "{:15.7E} {:15.7E} {:15.7E} {:15.7E}\n".format(
                            icoil.x[iseg], icoil.y[iseg], icoil.z[iseg], icoil.I
                        )
                    )
                wfile.write(
                    "{:15.7E} {:15.7E} {:15.7E} {:15.7E} {:} {:10} \n".format(
                        icoil.x[0], icoil.y[0], icoil.z[0], 0.0, icoil.group, icoil.name
                    )
                )
            wfile.write("end \n")
        return

    def save_gpec_coils(self, filename, split=True, nw=1, **kwargs):
        """Write the data in standard ascii format for GPEC

        Args:
            filename (str): path (if split==True) or file name to be saved.
            split (bool, optional): write each coil into a separate file. Defaults to True
            nw (integer, optional): number of windings. Defaults to 1.
        """
        if split:
            # write in independent files
            for icoil in list(self):
                with open(filename + icoil.name + ".dat", "w") as f:
                    # write the defining parameters
                    ncoil = 1
                    s = 1  # have to assume this?
                    nsec = len(icoil.x)
                    f.write(
                        "{:>5}{:>5}{:>5}{:8.2f}\n".format(ncoil, s, nsec, nw)
                    )  # the first line with periods
                    # write each coil x, y, z
                    for i in range(len(icoil.x)):
                        f.write(
                            "{:13.4e}{:13.4e}{:13.4e}\n".format(
                                icoil.x[i], icoil.y[i], icoil.z[i]
                            )
                        )
        else:
            # write into one file
            with open(filename, "w") as f:
                # write the defining parameters
                ncoil = len(self)
                s = 1  # have to assume this?
                nsec = len(self.data[0].x)
                f.write(
                    "{:>5}{:>5}{:>5}{:8.2f}\n".format(ncoil, s, nsec, nw)
                )  # the first line with periods
                # write each coil x, y, z
                for icoil in list(self):
                    for i in range(len(icoil.x)):
                        f.write(
                            "{:13.4e}{:13.4e}{:13.4e}\n".format(
                                icoil.x[i], icoil.y[i], icoil.z[i]
                            )
                        )
        return

    def toVTK(self, vtkname, line=True, height=0.1, width=0.1, **kwargs):
        """Write entire coil set into a VTK file

        Args:
            vtkname (str): VTK filename.
            line (bool, optional): Save coils as polylines or surfaces. Defaults to True.
            height (float, optional): Rectangle height when expanded to a finite cross-section. Defaults to 0.1.
            width (float, optional): Rectangle width when expanded to a finite cross-section. Defaults to 0.1.
            kwargs (dict): Optional kwargs passed to "polyLinesToVTK" or "meshio.Mesh.write".
        """
        from pyevtk.hl import polyLinesToVTK, gridToVTK

        if line:
            currents = []
            groups = []
            x = []
            y = []
            z = []
            lx = []
            for icoil in list(self):
                currents.append(icoil.I)
                groups.append(icoil.group)
                x.append(icoil.x)
                y.append(icoil.y)
                z.append(icoil.z)
                lx.append(len(icoil.x))
            kwargs.setdefault("cellData", {})
            kwargs["cellData"].setdefault("I", np.array(currents))
            kwargs["cellData"].setdefault("Igroup", np.array(groups))
            polyLinesToVTK(
                vtkname,
                np.concatenate(x),
                np.concatenate(y),
                np.concatenate(z),
                np.array(lx),
                **kwargs
            )
        else:
            import meshio

            points = []
            hedrs = []
            currents = []
            groups = []
            nums = []
            start = 0
            for i, icoil in enumerate(self):
                # example of meshio.Mesh can be found at https://github.com/nschloe/meshio
                xx, yy, zz = icoil.rectangle(width=width, height=height)
                xx = np.ravel(np.transpose(xx[0:4, :]))
                yy = np.ravel(np.transpose(yy[0:4, :]))
                zz = np.ravel(np.transpose(zz[0:4, :]))
                xyz = np.transpose([xx, yy, zz])
                points += xyz.tolist()
                # number of cells is npoints-1
                ncell = len(xx) // 4 - 1
                ind = np.reshape(np.arange(4 * ncell + 4) + start, (-1, 4))
                hedr = [list(ind[j, :]) + list(ind[j + 1, :]) for j in range(ncell)]
                hedrs += hedr
                currents += (icoil.I * np.ones(ncell)).tolist()
                groups += (icoil.group * np.ones(ncell, dtype=int)).tolist()
                nums += ((i + 1) * np.ones(ncell, dtype=int)).tolist()
                # update point index number
                start += len(xx)
            kwargs.setdefault("cell_data", {})
            # coil currents
            kwargs["cell_data"].setdefault("I", [currents])
            # current groups
            kwargs["cell_data"].setdefault("group", [groups])
            # coil index, starting from 1
            kwargs["cell_data"].setdefault("index", [nums])
            data = meshio.Mesh(points=points, cells=[("hexahedron", hedrs)], **kwargs)
            data.write(vtkname)
        return
